package be.iminds.ilabt.jfed.ui.javafx.util;

/**
 * TimeUtils  see http://stackoverflow.com/questions/3859288/how-to-calculate-time-ago-in-java
 */

import java.util.concurrent.TimeUnit;

import static java.util.concurrent.TimeUnit.MILLISECONDS;

public class TimeUtils {

    /**
     * Converts time to a human readable format within the specified range
     *
     * @param duration the time in milliseconds to be converted
     * @param max      the highest time unit of interest
     * @param min      the lowest time unit of interest
     */
    public static String formatMillis(long duration, TimeUnit max, TimeUnit min) {
        return formatMillis(duration, max, min, 0);
    }
    /**
     * Converts time to a human readable format within the specified range
     *
     * @param duration the time in milliseconds to be converted
     * @param max      the highest time unit of interest
     * @param min      the lowest time unit of interest
     * @param maxParts the maximum number of sub parts. 1 is a bit imprecise, 2 is nice, 3 is a lot. If 0, this is chosen automatically (typically 2 or 3).
     */
    public static String formatMillis(long duration, TimeUnit max, TimeUnit min, int maxParts) {
        StringBuilder res = new StringBuilder();

        TimeUnit current = max;

        int parts = 0;
        while (duration > 0) {
            long temp = current.convert(duration, MILLISECONDS);

            if (temp > 0) {
                duration -= current.toMillis(temp);
                res.append(temp).append(" ").append(current.name().toLowerCase());
                if (temp < 2) res.deleteCharAt(res.length() - 1);
                res.append(", ");
                parts++;
                if (maxParts > 0 && parts >= maxParts)
                    break;
            }

            if (current == min) break;

            current = TimeUnit.values()[current.ordinal() - 1];
        }

        // clean up our formatting....

        // we never got a hit, the time is lower than we care about
        if (res.lastIndexOf(", ") < 0) return "0 " + min.name().toLowerCase();

        // yank trailing  ", "
        res.deleteCharAt(res.length() - 2);

        //  convert last ", " to " and"
        int i = res.lastIndexOf(", ");
        if (i > 0) {
            res.deleteCharAt(i);
            res.insert(i, " and");
        }

        return res.toString();
    }
}