package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.lowlevel.credential.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.module.JFedWebApiClientModule;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTracker;
import be.iminds.ilabt.jfed.experiment.*;
import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig;
import be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupHelper;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedExperimenterGuiConfigProvider;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfigImpl;
import be.iminds.ilabt.jfed.experimenter_gui.config.UserInfoProvider;
import be.iminds.ilabt.jfed.git.SingleSshGitAuthPreferences;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.jobs.TestConnectivityJob;
import be.iminds.ilabt.jfed.highlevel.jobs.TestLinksJob;
import be.iminds.ilabt.jfed.highlevel.jobs.link_test.LinkTestListener;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.*;
import be.iminds.ilabt.jfed.util.lib.ConnectivityDetector;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.GuiConfigRSpecGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.GuiConfigStitchingTestRSpecGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.config.util.TestbedNodesMapsFetcher;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.util.ProxyServiceUtil;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.rspec.generator.StitchingTestRSpecGenerator;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.ui.cli2.instruction.*;
import be.iminds.ilabt.jfed.ui.cli2.logging.ActionLogger;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.name.Names;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.*;
import java.net.URL;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static be.iminds.ilabt.jfed.experiment.setup.ExperimentSetupConfig.ProjectSource.PROVIDED;

/**
 * InstructionExecutor
 */
public class InstructionExecutor {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(InstructionExecutor.class);
    protected static final ObjectMapper MAPPER = new ObjectMapper();

    @Nonnull private final Context context;
    @Nonnull private final TaskThread taskThread;
    @Nonnull private final HighLevelTaskFactory highLevelTaskFactory;
    @Nonnull private final Injector injector;

    public InstructionExecutor(@Nonnull Context context) {
        this.context = context;
        this.injector = context.getInjector();
        this.taskThread = injector.getInstance(TaskThread.class);
        this.highLevelTaskFactory = injector.getInstance(HighLevelTaskFactory.class);
    }

    /**
     * @param context
     * @param instruction
     * @return CLI exit status
     */
    public void run(@Nonnull Context context, @Nonnull Instruction instruction) throws CliExitException {
        LOG.debug("Executing "+instruction.getAction()+" instruction");

        if (instruction instanceof InstructionWithSlice && instruction.getAction() != Instruction.Action.CREATESLICE) {
            //all instructions that require recovering the slice as a first step
            InstructionWithSlice instructionWithSlice = (InstructionWithSlice) instruction;
            logUserInfo();

            LOG.debug("Getting Slice name and URN");
            String sliceName = instructionWithSlice.getSlice().getSliceName();
            if (sliceName == null)
                throw new ArgumentException("Slice name is not allowed to be null");
            GeniUrn sliceUrn = getSliceUrn(sliceName,
                    instructionWithSlice.getSlice().getProjectSource() == PROVIDED ?
                            instructionWithSlice.getSlice().getProject() : null);

            @Nonnull Experiment experiment = recoverExperiment(sliceUrn);
            @Nonnull Slice slice = experiment.getSlice();

            LOG.info("Recovered experiment, will now start it to finish recover. Experiment is now in state: "+experiment.getExperimentState());

            final ExperimentControllerFactory experimentControllerFactory =
                    injector.getInstance(ExperimentControllerFactory.class);
            ExperimentController experimentController = experimentControllerFactory.createExperimentController(experiment);

            CountDownLatch waitForExperimentRecovered = new CountDownLatch(1);
            experiment.addExperimentChangeListener(new ExperimentChangeListener() {
                @Override public void onExperimentStateChange(ExperimentState newExperimentState) {
                    if (newExperimentState != ExperimentState.RESTORING &&
                            newExperimentState != ExperimentState.PENDING )
                        waitForExperimentRecovered.countDown();
                }

                @Override public void onExperimentPartAdded(ExperimentPart experimentPart) { }
            });
            experimentController.start();

            try {
                if (!waitForExperimentRecovered.await(10, TimeUnit.SECONDS))
                    throw new CliExitException(Context.ExitReason.ERROR, "Wait for experiment recover took too long.");
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for experiment recover interrupted.");
            }

            LOG.info("Experiment is now in state: "+experiment.getExperimentState());

            assert experiment.getPartsSize() > 0;

            switch (instructionWithSlice.getAction()) {
                case DELETE: {
                    executeDelete(slice);
                    return;
                }
                case POA: {
                    //noinspection CastConflictsWithInstanceof
                    executePoa((PoaInstruction) instruction, slice);
                    return;
                }
                case LOGIN: {
                    //noinspection CastConflictsWithInstanceof
                    executeSshLogin((SshInstruction) instruction, slice);
                    return;
                }
                case SLICEINFO: {
                    //noinspection CastConflictsWithInstanceof
                    executeSliceInfo((SliceInfoInstruction) instruction, slice);
                    return;
                }
                case SLICEACTION: {
                    //noinspection CastConflictsWithInstanceof
                    executeSliceAction((SliceActionInstruction) instruction, slice,
                            experiment, experimentController);
                    return;
                }
                case MANIFEST: {
                    executeManifest(slice);
                    return;
                }
                case SLICECREDENTIAL: {
                    executeSliceCredential(slice);
                    return;
                }
                case RENEW: {
                    //noinspection CastConflictsWithInstanceof
                    executeRenew((RenewInstruction) instruction, slice, experiment);
                    return;
                }
                default: throw new CliExitException(Context.ExitReason.ERROR, "Action "+instruction.getAction()+" not implemented");
            }
        } else {
            switch (instruction.getAction()) {
                case LISTRESOURCES: {
                    //noinspection ConstantConditions
                    executeListResources((ListResourcesInstruction) instruction);
                    return;
                }
                case USERINFO: {
                    //noinspection ConstantConditions
                    executeUserInfo((UserInfoInstruction) instruction);
                    return;
                }
                case CREATESLICE: {
                    logUserInfo();
                    //noinspection ConstantConditions
                    CreateSliceInstruction createSliceInstruction = (CreateSliceInstruction) instruction;
                    executeCreateSlice(createSliceInstruction);
                    return;
                }
                case RUN: {
                    assert instruction instanceof RunInstruction;
                    //noinspection ConstantConditions
                    executeRun(context, (RunInstruction) instruction);
                    return;
                }
                default: throw new CliExitException(Context.ExitReason.ERROR, "Action "+instruction.getAction()+" not implemented");
            }

        }
        //unreachable
    }

    private void logUserInfo() {
        final GeniUser user = injector.getInstance(GeniUser.class);
        LOG.info("User " + user.getUserUrnString() + " will be used.");
        LOG.debug("User Authority: " + (user.getUserAuthorityServer() == null ? null :
                user.getUserAuthorityServer().getName()));
        LOG.debug("User Authority URN: " + user.getUserAuthorityServer().getDefaultComponentManagerUrn());
    }

    @Nonnull
    private GeniUrn getSliceUrn(@Nonnull String sliceName, @Nullable String projectName) throws CliExitException {
        //TODO allow slice URN to be specified directly
//        if (instruction.hasSliceUrn) {
//            //no need to retrieve URN
//        }

        ListSlicesTask listSlicesTask = highLevelTaskFactory.listSlices();
        TaskExecution listSlicesTaskExecution = taskThread.addTask(listSlicesTask);
        while (!listSlicesTaskExecution.isCompleted()) {
            try {
                Thread.sleep(100L);
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for list slices interrupted.");
            }
        }
        if (listSlicesTaskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            LOG.error("List Slices failed -> listSlicesTaskExecution.getState()="+listSlicesTaskExecution.getState(),
                    listSlicesTaskExecution.getException());
            throw new CliExitException(Context.ExitReason.BAD_ARGS, "List Slices failed");
        }
        GeniUrn sliceUrn = listSlicesTask.getSlices().stream()
                .filter(curSliceUrn -> curSliceUrn.getEncodedResourceName().equalsIgnoreCase(sliceName))
                .filter(curSliceUrn -> projectName == null || curSliceUrn.getEncodedTopLevelAuthority().toLowerCase().endsWith(":" + projectName.toLowerCase()))
                .findAny()
                .orElseThrow(() -> {
                    LOG.error("slice '"+sliceName+"' not found in "+listSlicesTask.getSlices());
//                    context.getActionLogger().errorPrintln("slice '"+sliceName+"' not found");
                    return new CliExitException(Context.ExitReason.BAD_ARGS, "slice '"+sliceName+"' not found");
                });
        return sliceUrn;
    }

    @Nonnull
    private Experiment recoverExperiment(@Nonnull GeniUrn sliceUrn) throws CliExitException {
        RecoverSliceTaskCliInteraction recoverSliceTaskInteraction = new RecoverSliceTaskCliInteraction(context);
        RecoverSliceTask recoverSliceTask = highLevelTaskFactory.createRecoverSliceTask(sliceUrn, recoverSliceTaskInteraction);
        recoverSliceTask.run();
        Slice slice;
        try {
            slice = recoverSliceTask.get();
            if (slice == null) {
                LOG.error("Failed to recover slice -> recoverSliceTask.get() == null");
                throw new CliExitException(Context.ExitReason.ERROR, "Failed to recover slice");
            }
        } catch (InterruptedException e) {
            //should not happen, as we run above
            LOG.error("Failed to recover slice -> InterruptedException where there shouldn't be one", e);
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to recover slice");
        } catch (ExecutionException e) {
            LOG.error("Failed to recover slice -> Error in RecoverSliceTask", e);
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to recover slice");
        }
        //note: recoverSliceTaskInteraction might not have experiment here! It gets it asynchronously.
        Experiment experiment = recoverSliceTask.getExperiment();
        if (experiment == null) {
            LOG.error("Failed to recover slice -> recoverSliceTaskInteraction.getExperiment() == null");
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to recover slice");
        }
        return experiment;
    }

    private void executeDelete(@Nonnull Slice slice) throws CliExitException {
        LOG.debug("executeDelete "+slice.getUrnString());
        List<TaskExecution> taskExecutions = new ArrayList<>();
        for (Server server : slice.getConnectAuthorities()) {
            DeleteSliversAtAuthorityTask deleteTask = highLevelTaskFactory.deleteSliversAtAuthority(slice, server);
            taskExecutions.add(taskThread.addTask(deleteTask));
        }
        for (TaskExecution taskExecution: taskExecutions) {
            while (!taskExecution.isCompleted()) {
                try {
                    Thread.sleep(100L);
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for delete interrupted.");
                }
            }
            if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                LOG.error("A Delete failed", taskExecution.getException());
                throw new CliExitException(Context.ExitReason.ERROR, "A Delete failed.");
            }
        }
        LOG.debug("Delete finished.");
        context.getActionLogger().actionPrintln("Deleted: "+slice.getUrnString());
    }


    private void executeSshLogin(@Nonnull SshInstruction instruction,
                            @Nonnull Slice slice) throws CliExitException {
        throw new RuntimeException("This action is not yet implemented");
    }

    private void executePoa(@Nonnull PoaInstruction instruction,
                            @Nonnull Slice slice) throws CliExitException {
        LOG.debug("executePoa "+slice.getUrnString());
        PoaInstruction.PoaAction poaAction = instruction.getPoaAction();

        GeniUrn sliverUrn;
        Sliver sliver;
        try {
            if (instruction.getSliverUrn() != null) {
                sliverUrn = new GeniUrn(instruction.getSliverUrn());
                sliver = slice.findSliver(sliverUrn);
                if (sliver == null)
                    throw new CliExitException(Context.ExitReason.ERROR, "Sliver not found: "+sliverUrn);
            }
            else {
                sliverUrn = null;
                sliver = null;
            }
        } catch (GeniUrn.GeniUrnParseException e) {
            throw new ArgumentException("POA sliver URN is invalid: '"+instruction.getSliverUrn()+"'", e);
        }

        List<TaskExecution> taskExecutions = new ArrayList<>();
        if (sliver != null) {
            Task poaTask;
            switch (poaAction) {
                case Restart: {
                    poaTask = highLevelTaskFactory.restart(slice, sliverUrn, sliver.getServer());
                    break;
                }
                case ReloadOs: {
                    poaTask = highLevelTaskFactory.reloadOS(slice, sliverUrn, sliver.getServer());
                    break;
                }
                case ConsoleUrl: {
                    poaTask = highLevelTaskFactory.openConsole(sliver);
                    break;
                }
                default: throw new ArgumentException("Unknown POA Action: '"+poaAction+"'");
            }
            taskExecutions.add(taskThread.addTask(poaTask));
        } else {
            switch (poaAction) {
                case Restart: {
                    for (Server server : slice.getConnectAuthorities()) {
                        Task poaTask = highLevelTaskFactory.restartSlice(slice, server);
                        taskExecutions.add(taskThread.addTask(poaTask));
                    }
                    break;
                }
                case ReloadOs: {
                    for (Sliver s : slice.getSliversCopy()) {
                        Task poaTask = highLevelTaskFactory.reloadOS(s);
                        taskExecutions.add(taskThread.addTask(poaTask));
                    }
                    break;
                }
                case ConsoleUrl: {
                    throw new ArgumentException("geni_console_url requires a sliver URN");
                }
                default:
                    throw new ArgumentException("Unknown POA Action: '" + poaAction + "'");
            }

        }
        for (TaskExecution taskExecution: taskExecutions) {
            while (!taskExecution.isCompleted()) {
                try {
                    Thread.sleep(100L);
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for POA interrupted.");
                }
            }
            if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                LOG.error("One POA failed", taskExecution.getException());
                throw new CliExitException(Context.ExitReason.ERROR, "One POA failed.");
            }
        }
        LOG.debug("POA finished.");
        context.getActionLogger().actionPrintln("POA "+poaAction+": "+slice.getUrnString());
    }

    private void executeSliceAction(@Nonnull SliceActionInstruction instruction,
                                    @Nonnull Slice slice,
                                    @Nonnull Experiment experiment,
                                    @Nonnull ExperimentController experimentController) throws CliExitException {
        LOG.debug("executeSliceAction "+slice.getUrnString());
        SliceActionInstruction.SliceAction sliceAction = instruction.getSliceAction();

        if (sliceAction == SliceActionInstruction.SliceAction.RerunESpec && instruction.getEspec() == null) {
            throw new ArgumentException("SliceAction "+sliceAction+" requires espec argument");
        }

        if (sliceAction != SliceActionInstruction.SliceAction.RerunESpec && instruction.getEspec() != null) {
            throw new ArgumentException("SliceAction "+sliceAction+" makes no sense with espec argument");
        }

        switch (sliceAction) {
            case LinkTest: {
                experimentController.setExperimentLinkTesterFactory((exp) -> {
                    //noinspection ObjectEquality
                    assert exp == experiment;
                    return new TestLinksJob(
                            exp,
                            injector.getInstance(HighLevelTaskFactory.class),
                            injector.getInstance(TaskThread.class),
                            injector.getInstance(GeniUserProvider.class),
                            injector.getInstance(JFedPreferences.class),
                            injector.getInstance(ProxyPreferencesManager.class),
                            injector.getInstance(ProxyServiceUtil.class),
                            injector.getInstance(JobStateFactory.class),
                            injector.getInstance(ProxySocketFactoryProvider.class));});
                LinkTestResultToOutput linkTestLogger = new LinkTestResultToOutput(context.getActionLogger());
                experiment.addLinkTestListener(linkTestLogger);
                assert experiment.getPartsSize() > 0;
                CompletableFuture<Boolean> future = experimentController.testLinks();
                try {
                    Boolean testSuccess = future.get();
//                    if (!success) {
//                        throw new CliExitException(Context.ExitReason.ERROR, "Link test failed");
//                    }
                    boolean linksOk = linkTestLogger.getWorstResult() != null &&
                            linkTestLogger.getWorstResult() == TestLinksJob.LinkTestResult.SUCCESS;
                    LOG.debug("Link Test finished.");
                    Map<String, Object> jsonRes = new HashMap<>();
                    jsonRes.put("All Links OK", linksOk);
                    List<Object> jsonReports = new ArrayList<>();
                    if (linkTestLogger.reports != null) {
                        for (TestLinksJob.LinkTestReport linkTestReport: linkTestLogger.reports) {
                            Map<String, Object> jsonReport = new HashMap<>();
                            jsonReport.put("from", linkTestReport.getFrom().getClientId());
                            jsonReport.put("to", linkTestReport.getTo().getClientId());
                            jsonReport.put("summary", linkTestReport.getSummary());
                            jsonReport.put("expected_mbps", linkTestReport.getExpectedLinkSpeed_mbps());
                            jsonReport.put("actual_throughput_mbps", linkTestReport.getActualThroughput_mbps());
                            jsonReport.put("actual_linkspeedsetting_mbps", linkTestReport.getActualLinkSpeedSetting_mbps());
                            jsonReport.put("loss_percent", linkTestReport.getActualLoss_percent());
                            jsonReports.add(jsonReport);
                        }
                    }
                    if (!jsonReports.isEmpty())
                        jsonRes.put("reports", jsonReports);
                    context.getActionLogger().actionJsonOut(jsonRes);
                    context.getActionLogger().actionPrintln("All Links OK: "+linksOk);
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for test links interrupted.");
                } catch (ExecutionException e) {
                    LOG.error("Failed to test links -> Error in experimentController.testLinks", e);

                    throw new CliExitException(Context.ExitReason.ERROR, "Failed to test links");
                }
                break;
            }
            case RerunESpec: {
                ESpecLoggingToOutput especListener =
                        new ESpecLoggingToOutput(context.getActionLogger());
                experiment.addESpecLogListener(especListener);
                ESpecLoggingCounter eSpecLoggingCounter = new ESpecLoggingCounter();
                experiment.addESpecLogListener(eSpecLoggingCounter);

                try {
                    final GeniUser user = injector.getInstance(GeniUser.class);
                    ESpecBundle experimentSpecificationBundle = ExperimentSetupHelper.createESpecBundle(instruction.getEspec(), user);
                    ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
                    ExperimentSpecification experimentSpecification = parser.parse(experimentSpecificationBundle.getExperimentSpecificationYml());
                    experiment.resetExperimentSpecification(
                            experimentSpecification,
                            experimentSpecificationBundle,
                            getRSpecGeneratorFactory(injector),
                            new SingleSshGitAuthPreferences(user.getPrivateKey()));
                } catch (IOException | ESpecBundle.ESpecBundleInitException | ExperimentSpecificationParser.ExperimentSpecificationParseException e) {
                    LOG.error("ESpec Fetch failed", e);
                    throw new CliExitException(Context.ExitReason.ERROR, "ESpec Fetch failed", e);
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for ESpec Fetch interrupted.");
                }

                if (experiment.getExperimentSpecificationFileManager() == null) {
                    throw new CliExitException(Context.ExitReason.ERROR, "ESpec setup failed: experiment.experimentSpecificationFileManager == null");
                }
                experiment.getExperimentSpecificationFileManager().fetchAll();
                if (experiment.getExperimentSpecificationFileManager().getErrorOccured())
                    throw new CliExitException(Context.ExitReason.ERROR, "Error fetching ESpec resources");

                CompletableFuture<Boolean> future = experimentController.rerunESpec(null, null);
                try {
                    Boolean success = future.get();

                    LOG.debug("Rerun ESpec finished. success="+success);

                    Map<String, Object> jsonRes = new HashMap<>();
                    jsonRes.put("Overview Counts", eSpecLoggingCounter.getCounts());
                    jsonRes.put("Rerun ESpec Success", success);
                    context.getActionLogger().actionJsonOut(jsonRes);
                    if (!success) {
                        throw new CliExitException(Context.ExitReason.ERROR, "Rerun ESpec failed");
                    }
                    context.getActionLogger().actionPrintln("Reran ESpec");
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for ESpec Rerun interrupted.");
                } catch (ExecutionException e) {
                    LOG.error("Failed to Rerun ESpec -> Error in experimentController.rerunESpec", e);

                    throw new CliExitException(Context.ExitReason.ERROR, "Failed to Rerun ESpec");
                }
                break;
            }
            case ConnectivityTest: {
                experimentController.setExperimentConnectivityTesterFactory((exp) -> new TestConnectivityJob(
                        exp,
                        injector.getInstance(HighLevelTaskFactory.class),
                        injector.getInstance(TaskThread.class),
                        injector.getInstance(GeniUserProvider.class),
                        injector.getInstance(JFedPreferences.class),
                        injector.getInstance(ProxyPreferencesManager.class),
                        injector.getInstance(ProxyServiceUtil.class)));
                //TODO not very clean with this cast
                ExperimentControllerImpl experimentControllerImpl = (ExperimentControllerImpl) experimentController;
                CompletableFuture<TestConnectivityJob.TestConnectivityJobResult> future = experimentControllerImpl.testConnectivity();
                try {
                    TestConnectivityJob.TestConnectivityJobResult jobres = future.get();

                    LOG.debug("Connectivity Test finished.");
                    Map<String, Object> connectivityTestResult = new HashMap<>();
                    connectivityTestResult.put("All Connectivity OK", jobres.isAllSuccessFull());
//                    if (!jobres.getFailedNodes().isEmpty()) {
                    connectivityTestResult.put("Failed Nodes", jobres.getFailedNodes().stream().map(FXRspecNode::getClientId).collect(Collectors.toList()));
//                    }
//                    if (!jobres.getSuccessfulNodes().isEmpty()) {
                    connectivityTestResult.put("Successful Nodes", jobres.getSuccessfulNodes().stream().map(FXRspecNode::getClientId).collect(Collectors.toList()));
//                    }
                    context.getActionLogger().actionJsonOut(connectivityTestResult);
                    context.getActionLogger().actionPrintln("Connectivity Test success: "+jobres.isAllSuccessFull());
                } catch (InterruptedException e) {
                    throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for Connectivity test interrupted.");
                } catch (ExecutionException e) {
                    LOG.error("Failed to test connectivity -> Error in experimentController.testConnectivity", e);

                    throw new CliExitException(Context.ExitReason.ERROR, "Failed to test connectivity");
                }
                break;
            }
            default:
                throw new ArgumentException("Unhandled Slice Action: '" + sliceAction + "'");
        }

//        LOG.debug("Slice Action finished.");
//        context.getActionLogger().actionPrintln("Finished Slice Action "+sliceAction+": "+slice.getUrnString());
    }

    private void executeManifest(@Nonnull Slice slice) throws CliExitException {
        LOG.debug("executeManifest "+slice.getUrnString());
        ManifestRspecSource manifestRspecSource = slice.getManifestRspec();
        String manifestRspec = manifestRspecSource == null ? null : manifestRspecSource.getRspecXmlString();
        if (manifestRspec != null)
            context.getActionLogger().actionPrint(manifestRspec);
        else
            context.getActionLogger().errorPrintln("No manifest found");
        LOG.debug("Manifest returned");
    }

    @SuppressWarnings("UseOfObsoleteDateTimeApi")
    private void executeSliceInfo(@Nonnull SliceInfoInstruction instruction, @Nonnull Slice slice) throws CliExitException {
        LOG.debug("executeSliceInfo "+slice.getUrnString());

        Map<String, Object> sliceInfoMap = new HashMap<>();

        sliceInfoMap.put("sliceUrn", slice.getUrnString());

        if (instruction.isShowMetaInfo()) {
            final GeniUser user = injector.getInstance(GeniUser.class);
            sliceInfoMap.put("userUrn", user.getUserUrnString());

            sliceInfoMap.put("usedSpeaksfor", false);
//            sliceInfoMap.put("usedSpeaksfor", usesSpeaksFor);
//            sliceInfoMap.put("speaksForCredential", speaksForCredentials.get(0));
//            sliceInfoMap.put("speakingForUserUrn", speakingForUser);

            sliceInfoMap.put("sliceAggregates", slice.getAuthorities().stream().map(Server::getDefaultComponentManagerUrn).collect(Collectors.toList()));
            Instant sliceExpire = slice.getExpirationDate();
            sliceInfoMap.put("sliceExpiration", sliceExpire == null ? null : RFC3339Util.instantToRFC3339String(sliceExpire));

        }
        if (instruction.isShowUsers()) {
            sliceInfoMap.put("sliceUsers", slice.getUsers().stream().map(GeniUrn::toString).collect(Collectors.toList()));
        }
        if (instruction.isShowCredentials()) {
//            sliceInfoMap.put("userCredentials", todo);
            sliceInfoMap.put("sliceCredentials", slice.getCredentials().stream().map(AnyCredential::getCredentialXml).collect(Collectors.toList()));
        }


        context.getActionLogger().actionPrintln("Slice URN: " + sliceInfoMap.get("sliceUrn"));
        if (sliceInfoMap.get("userUrn") != null)
            context.getActionLogger().actionPrintln("User URN: " + sliceInfoMap.get("userUrn"));
//        if (sliceInfoMap.get("userCredentials") != null)
//            context.getActionLogger().actionPrintln("User Credentials: "+sliceInfoMap.get("userCredentials"));
        if (sliceInfoMap.get("sliceExpiration") != null)
            context.getActionLogger().actionPrintln("Slice Expiration: " + sliceInfoMap.get("sliceExpiration"));


        if (instruction.isShowStatus()) {
            final Map<String, Object> outerStatusMap = new HashMap<>();
            final Map<String, Object> outerStatusAmDetailMap = new HashMap<>();
            outerStatusMap.put("AMs", outerStatusAmDetailMap);
            sliceInfoMap.put("status", outerStatusMap);

            for (Server amServer: slice.getConnectAuthorities()) {
                StatusDetails statusDetailsForAM = slice.getSliversStream()
                        .filter(s -> s.getServer().equals(amServer))
                        .map(Sliver::getStatus)
                        .filter(Objects::nonNull)
                        .reduce(StatusDetails.identityStatus, StatusDetails.merge);

                context.getActionLogger().actionPrintln(
                        "Sliver at  " + amServer.getDefaultComponentManagerUrn() + " has status " + statusDetailsForAM.getGlobalStatus());

                Map<String, Object> outerAMHt = new HashMap<>();
                outerAMHt.put("amUrn", amServer.getDefaultComponentManagerUrn());
                outerAMHt.put("amGlobalSliverStatus", statusDetailsForAM.getGlobalStatus());
                Date earliestAMSliverExpireDate = statusDetailsForAM.getEarliestKnownExpires();
                if (earliestAMSliverExpireDate != null)
                    outerAMHt.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestAMSliverExpireDate));
                List<Map<String, Object>> detailStatus = new ArrayList<>();
                for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetailsForAM.getStatusBySliverUrn().entrySet()) {
                    Map<String, Object> detailHt = new HashMap<>();
                    detailHt.put("urn", detail.getKey()+"");
                    detailHt.put("type", "sliver");
                    detailHt.put("status", detail.getValue() + "");
                    detailStatus.add(detailHt);
                }
                for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetailsForAM.getStatusByComponentUrn().entrySet()) {
                    Map<String, Object> detailHt = new HashMap<>();
                    detailHt.put("urn", detail.getKey()+"");
                    detailHt.put("type", "component");
                    detailHt.put("status", detail.getValue() + "");
                    detailStatus.add(detailHt);
                }
                outerAMHt.put("details", detailStatus);
                outerStatusAmDetailMap.put(amServer.getDefaultComponentManagerUrn(), outerAMHt);
            }


            Date earliestSliverExpireDate = slice.getStatus() == null ? null : slice.getStatus().getEarliestKnownExpires();
            if (earliestSliverExpireDate != null)
                outerStatusMap.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestSliverExpireDate));

            context.getActionLogger().actionPrintln("Status Info:");
            if (outerStatusMap.get("earliestSliverExpireDate") != null)
                context.getActionLogger().actionPrintln("   Earliest sliver expire date: " + outerStatusMap.get("earliestSliverExpireDate"));
            else
                context.getActionLogger().actionPrintln("   Earliest sliver expire date: no slivers");
            Map<String, Object> amsHt = (Map<String, Object>) outerStatusMap.get("AMs");
            Set<Map.Entry<String, Object>> entries = amsHt.entrySet();
            for (Map.Entry e : entries) {
                String key = (String) e.getKey();
                Map amHt = (Map) e.getValue();

                String amUrn = amHt.get("amUrn") + "";
                String amGlobalSliverStatus = amHt.get("amGlobalSliverStatus") + "";
                String earliestAMSliverExpireDateStr = amHt.get("earliestSliverExpireDate") == null ? null : amHt.get("earliestSliverExpireDate") + "";

                context.getActionLogger().actionPrintln("   AM: " + amUrn + " " + amGlobalSliverStatus + (earliestAMSliverExpireDateStr == null ? "" : " " + earliestAMSliverExpireDateStr));

                List details = (List) amHt.get("details");
                for (Object detail : details) {
                    Map detailHt = (Map) detail;
                    String urn = detailHt.get("urn") + "";
                    String type = (String) detailHt.get("type");
                    String status = detailHt.get("status") + "";
                    context.getActionLogger().actionPrintln("      " + type + ": " + urn + " " + status);
                }
            }
        }

        if (sliceInfoMap.get("sliceCredentials") != null) {
            List sliceCreds = (List) sliceInfoMap.get("sliceCredentials");
            int i = 0;
            if (sliceCreds != null)
                for (Object c : sliceCreds) {
                    if (c instanceof AnyCredential) {
                        AnyCredential ac = (AnyCredential) c;
                        context.getActionLogger().actionPrintln("Slice Credential " + (++i) + ":\n" + ac.getCredentialXml() + "\n\nEND SLICE CREDENTIAL\n");
                    }
                }
        }

        context.getActionLogger().actionJsonOut(sliceInfoMap);

        LOG.debug("Slice Info returned");
    }

    private void executeRenew(@Nonnull RenewInstruction instruction,
                              @Nonnull Slice slice,
                              @Nonnull Experiment experiment) throws CliExitException {
        LOG.debug("executeRenew "+slice.getUrnString());

        final ExperimentControllerFactory experimentControllerFactory =
                injector.getInstance(ExperimentControllerFactory.class);

        ExperimentController experimentController = experimentControllerFactory.createExperimentController(experiment);

        Instant requestedRenewInstant = Instant.now().plus(instruction.getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
        CompletableFuture<ExperimentTaskStatus> future = experimentController.renew(requestedRenewInstant);

        try {
            ExperimentTaskStatus renewStatus = future.get();

            Map<String, Object> jsonOut = new HashMap<>();
            jsonOut.put("SliceUrn", slice.getUrnString());
            jsonOut.put("Renew Status", renewStatus.name());
            jsonOut.put("Renew Success", renewStatus == ExperimentTaskStatus.SUCCESS);
            context.getActionLogger().actionJsonOut(jsonOut);

            if (renewStatus != ExperimentTaskStatus.SUCCESS) {
                LOG.error("Renew failed: "+renewStatus);
                throw new CliExitException(Context.ExitReason.ERROR, "Failed to Renew slice");
            }

            LOG.debug("Renew finished.");
            context.getActionLogger().actionPrintln("Renewed: "+slice.getUrnString());
        } catch (InterruptedException e) {
            throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for Renew interrupted.");
        } catch (ExecutionException e) {
            LOG.error("Failed to renew slice -> Error in experimentController.renew", e);

            Map<String, Object> jsonOut = new HashMap<>();
            jsonOut.put("SliceUrn", slice.getUrnString());
            jsonOut.put("Renew Error", e.getMessage());
            jsonOut.put("Renew Success", false);
            context.getActionLogger().actionJsonOut(jsonOut);

            throw new CliExitException(Context.ExitReason.ERROR, "Failed to Renew slice");
        }
    }

    @Nonnull
    private List<String> getUserProjects(String taskDesc) throws CliExitException {
        ListSubAuthoritiesTask listSubAuthoritiesTask = highLevelTaskFactory.listSubAuthorities();
        TaskExecution taskExecution = taskThread.addTask(listSubAuthoritiesTask);

        while (!taskExecution.isCompleted()) {
            try {
                Thread.sleep(100L);
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for "+taskDesc+" interrupted.");
            }
        }
        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            LOG.error("Failed to "+taskDesc, taskExecution.getException());
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to "+taskDesc);
        }
        if (listSubAuthoritiesTask.getSubAuthNames() == null)
            return Collections.emptyList();
        return listSubAuthoritiesTask.getSubAuthNames();
    }

    @Nonnull
    private List<GeniUrn> getUserSlices(String taskDesc) throws CliExitException {
        ListSlicesTask listSlicesTask = highLevelTaskFactory.listSlices();
        TaskExecution taskExecution = taskThread.addTask(listSlicesTask);

        while (!taskExecution.isCompleted()) {
            try {
                Thread.sleep(100L);
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for "+taskDesc+" interrupted.");
            }
        }
        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            LOG.error("Failed to "+taskDesc, taskExecution.getException());
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to "+taskDesc);
        }
        if (listSlicesTask.getSlices() == null)
            return Collections.emptyList();
        return listSlicesTask.getSlices();
    }

    private void executeCreateSlice(@Nonnull CreateSliceInstruction instruction) throws CliExitException {
        LOG.debug("executeCreateSlice");

        LOG.debug("Getting Slice name and Project");
        String sliceName = instruction.getSlice().getSliceName();
        if (sliceName == null)
            throw new ArgumentException("Slice name is not allowed to be null");

        String projectName;
        switch (instruction.getSlice().getProjectSource()) {
            case AUTO: {
                @Nonnull List<String> projectNames = getUserProjects("Create Slice");
                if (projectNames.isEmpty()) {
                    LOG.error("Failed to get a list of user projects");
                    throw new CliExitException(Context.ExitReason.ERROR, "Failed to get a list of projects");
                }
                projectName = projectNames.get(0);
                LOG.debug("Choosing Project "+projectName+" from "+projectNames);
                break;
            }
            case PROVIDED: {
                projectName = instruction.getSlice().getProject();
                if (projectName == null)
                    throw new ArgumentException("No project provided to create slice in");
                break;
            }
            case NONE: {
                projectName = null;
                break;
            }
            default: throw new CliExitException(Context.ExitReason.ERROR, "unsupported project source: "+instruction.getSlice().getProjectSource());
        }

        Instant requestedEndTime = Instant.now().plus(instruction.getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
        List<GeniUrn> initialShareWithUsers = Collections.emptyList();

        CreateSliceTask createSliceTask = highLevelTaskFactory.createSlice(sliceName, projectName, initialShareWithUsers, requestedEndTime);
        LOG.debug("Creating slice "+sliceName+" in project "+projectName);
        TaskExecution taskExecution = taskThread.addTask(createSliceTask);

        while (!taskExecution.isCompleted()) {
            try {
                Thread.sleep(100L);
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.INTERRUPTED, "Wait for create slice interrupted.");
            }
        }
        if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
            LOG.error("Failed to Create slice", taskExecution.getException());
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to Create slice");
        }
    }

    private void executeSliceCredential(@Nonnull Slice slice) throws CliExitException {
        LOG.debug("executeSliceCredential "+slice.getUrnString());
        context.getActionLogger().actionJsonOut(
                slice.getCredentials().stream()
                        .map(AnyCredential::getCredentialXml)
                        .collect(Collectors.toList())
        );
        context.getActionLogger().actionPrintln(
                slice.getCredentials().stream()
                        .map(AnyCredential::getCredentialXml)
                        .collect(Collectors.joining("\n\n"))
        );
    }


    private void executeListResources(@Nonnull ListResourcesInstruction instruction) throws CliExitException {
        throw new RuntimeException("This action is not yet implemented");
    }

    private void executeUserInfo(@Nonnull UserInfoInstruction instruction) throws CliExitException {
        LOG.debug("executeUserInfo");

        final GeniUser user = injector.getInstance(GeniUser.class);
        String authUrn = user.getUserAuthorityServer() == null ? null :
                user.getUserAuthorityServer().getDefaultComponentManagerUrn();
        String authName = user.getUserAuthorityServer() == null ? null :
                user.getUserAuthorityServer().getName();

        @Nonnull List<String> projectNames = getUserProjects("fetch user project info");
        if (projectNames.isEmpty()) {
            LOG.error("Failed to get a list of user projects");
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to get a list of projects");
        }

        @Nonnull List<GeniUrn> sliceUrns = getUserSlices("fetch user slice info");
        if (sliceUrns.isEmpty()) {
            LOG.error("Failed to get a list of user slices");
            throw new CliExitException(Context.ExitReason.ERROR, "Failed to get a list of slices");
        }

        Map<String,Object> userInfoMap = new HashMap<>();
        userInfoMap.put("urn", user.getUserUrnString());
        userInfoMap.put("auth_urn", authUrn);
        userInfoMap.put("auth_name", authName);
        userInfoMap.put("slices", sliceUrns.stream().map(GeniUrn::toString).collect(Collectors.toList()));
        userInfoMap.put("projects", projectNames);
        context.getActionLogger().actionJsonOut(userInfoMap);

        context.getActionLogger().actionPrintln("User Info: ");
        context.getActionLogger().actionPrintln("   Urn: "+user.getUserUrnString());
        context.getActionLogger().actionPrintln("   Auth Urn: "+authUrn);
        context.getActionLogger().actionPrintln("   Auth Name: "+authName);
        context.getActionLogger().actionPrintln("   Projects: "+ projectNames.stream().collect(Collectors.joining(",")));
        context.getActionLogger().actionPrintln("   Slices: "+ sliceUrns.stream().map(GeniUrn::toString).collect(Collectors.joining(",")));
    }

    private void executeRun(@Nonnull Context context, @Nonnull RunInstruction instruction) throws CliExitException {
        LOG.debug("Executing RunInstruction");

        final Injector injector = context.getInjector();

        //TODO: add support for first fetching advertisement
        final AdvertisementRspecSource advertisementRspecSource = null;

        final GeniUrn bindRspecUrn;
        if (instruction.getExperiment().getRequestRSpec() != null &&
                instruction.getExperiment().getRequestRSpec().getSource().isProvide()) {
            ExperimentSetupConfig.ProvidedRequestRSpec pReqRspec =
                    (ExperimentSetupConfig.ProvidedRequestRSpec) instruction.getExperiment().getRequestRSpec();
            if (pReqRspec.getBindUnboundNodesUrn() != null) {
                bindRspecUrn = verifyBindRspecUrn(
                        context,
                        pReqRspec.getBindUnboundNodesUrn(),
                        injector.getInstance(AuthorityFinder.class));
                context.throwIfExited();
            }
            else
                bindRspecUrn = null;
        } else {
            bindRspecUrn = null;
        }

        LOG.debug("Setting up ExperimentSetupHelper");
        final ExperimentSetupHelper experimentSetupHelper = new ExperimentSetupHelper(
                "cli",

                context.getExperimentSetupLogger(),
                instruction.getExperiment(),
                null,

                () -> getRSpecGeneratorFactory(injector),
                advertisementRspecSource,

                injector
        );

        try {
            LOG.debug("Fetching RSpec");
            experimentSetupHelper.getRspec(bindRspecUrn == null ? null : bindRspecUrn.toString());
        } catch (Exception e) {
            LOG.error("Exception while retrieving request RSpec", e);
            context.exitWithCode(Context.ExitReason.ERROR, "Exception while retrieving request RSpec: "+e.getMessage());
        }
        context.throwIfExited();


//        if (context.getGenericCliArguments().isStitching() && experimentSetupHelper.getContext().requestRspecSource != null) {
//            if (!experimentSetupHelper.getContext().requestRspecSource.getStringRspec().isStitching(testbedInfoSource, authorityFinder)) {
//                context.exitWithCode(Context.ExitReason.BAD_ARGS, "The --stitching option was set, but the rspec does not specify any stitched links: Aborted CreateSliver.");
//                context.throwIfExited();
//            }
//        }

        logUserInfo();

        LOG.debug("Finding Project");
        experimentSetupHelper.findProject();


        LOG.debug("Getting slice name and SSH Keys");
        String sliceName = instruction.getExperiment().getSlice().getSliceName();
        if (sliceName == null)
            throw new ArgumentException("Slice name is not allowed to be null");

        List<UserSpec> userSpecs = getUserSpecs(instruction.getSshKeys());

        LOG.debug("Creating Experiment");
        experimentSetupHelper.createExperiment(sliceName, userSpecs,
                !instruction.getSlice().getFailOnNonExistingSlice(), !instruction.getSlice().getFailOnExistingSlice());

        //handle espec logging
        if (instruction.getExperiment().getESpec() != null) {
            ESpecLoggingToOutput especListener =
                    new ESpecLoggingToOutput(context.getActionLogger());
            experimentSetupHelper.getContext().experiment.addESpecLogListener(especListener);
        }

        //handle connectivity test logging
        if (instruction.getExperiment().getRunLinkTest()) {
            LinkTestResultToOutput linkTestListener =
                    new LinkTestResultToOutput(context.getActionLogger());
            experimentSetupHelper.getContext().experiment.addLinkTestListener(linkTestListener);
        }


        LOG.debug("Starting Experiment");
        experimentSetupHelper.startExperiment(injector);



        LOG.debug("Waiting for Experiment readiness");
        while (experimentSetupHelper.mustWaitForExperiment()) {
            try {
                Thread.sleep(1000L);
            } catch (InterruptedException e) {
                throw new CliExitException(Context.ExitReason.ERROR, "Interrupted");
            }
        }
        LOG.debug("Experiment ready wait completed");

//
////            Map sliceInfoHT = new HashMap();
//
////SF//             boolean usesSpeaksFor = cli.speaksForArgs != null && cli.speaksForArgs.length > 0;
////SF//             sliceInfoHT.put("usedSpeaksfor", usesSpeaksFor);
////SF//
////SF//             GeniUrn speakingForUser = null;
////SF//             List<AnyCredential> speaksForCredentials = new ArrayList<>();
////SF//             if (usesSpeaksFor) {
////SF//                 LOG.debug("There are "+cli.speaksForArgs.length+" speaksfor args: "+Arrays.asList(cli.speaksForArgs));
////SF//                 for (String speaksForArg : cli.speaksForArgs) {
////SF//                     List<String> speaksForCredentialStrings;
////SF//                     try {
////SF//                         String speaksForMultiCredentialString = IOUtils.fileToString(speaksForArg);
////SF//
////SF//                         speaksForCredentialStrings = extractCredentials(speaksForMultiCredentialString);
////SF//                     } catch (IOException e) {
////SF//                         LOG.error("Error reading speaksFor credential from file \"" + speaksForArg + "\"", e);
////SF//                         err.println("Error reading speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
////SF//                         return 3;
////SF//                     }
////SF//                     LOG.debug("File \""+speaksForArg+"\" contains "+speaksForCredentialStrings.size()+" credentials.");
////SF//                     for (String speaksForCredentialString : speaksForCredentialStrings) {
////SF//                         AnyCredential speaksForCredential;
////SF//                         GeniUrn currentSpeakingForUser = null;
////SF//                         try {
////SF//                             speaksForCredential = AnyCredential.createAutoDetect("speaksFor", speaksForCredentialString);
////SF//                         } catch (CredentialException e) {
////SF//                             LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"", e);
////SF//                             err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
////SF//                             return 3;
////SF//                         }
////SF//                         if (speaksForCredential == null) {
////SF//                             LOG.error("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
////SF//                             err.println("Error parsing speaksFor credential from file \"" + speaksForArg + "\"");
////SF//                             return 3;
////SF//                         }
////SF//                         if (!speaksForCredential.isSpeaksFor()) {
////SF//                             LOG.warn("WARNING: parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential");
////SF//                             err.println("WARNING parsed speaksFor credential from file \"" + speaksForArg + "\" is NOT a speaksFor credential. (Will continue anyway)");
////SF//                         }
////SF//                         try {
////SF//                             boolean supportedCredType = false;
////SF//                             if (speaksForCredential instanceof AbacCredential) {
////SF//                                 supportedCredType = true;
////SF//                                 AbacCredential c = (AbacCredential) speaksForCredential;
////SF//                                 if (!c.isSpeaksFor()) {
////SF//                                     LOG.error("Provided ABAC credential is not a speaksfor credential");
////SF//                                     err.println("Provided ABAC credential is not a speaksfor credential");
////SF//                                     return 3;
////SF//                                 }
////SF//                                 currentSpeakingForUser = c.getSpokenForUrn();
////SF//                                 if (c.getSpokenForUrn() != null) {
////SF//                                     if (speakingForUser != null) {
////SF//                                         LOG.debug("Speaksfor chain search: speakingForUser="+speakingForUser);
////SF//                                         LOG.debug("Speaksfor chain search: checking "+speaksForCredentials.size()+" credentials");
////SF//                                         //there are multiple abac credentials, so it's harder to find who is spoken for
////SF//                                         //assume all credentials are ABAC, and start complex processing of chain
////SF//                                         Map<String, GeniUrn> spokenForByKeyId = new HashMap<>();
////SF//                                         List<String> speakersKeyIds = new ArrayList<>();
////SF//
////SF//                                         List<AnyCredential> allCurrentlyKnownSpeaksForCredentials = new ArrayList<>(speaksForCredentials);
////SF//                                         allCurrentlyKnownSpeaksForCredentials.add(speaksForCredential);
////SF//                                         for (AnyCredential anyc : allCurrentlyKnownSpeaksForCredentials) {
////SF//                                             AbacCredential abacc = (AbacCredential) anyc;
////SF//                                             assert !spokenForByKeyId.containsKey(abacc.getHeadKeyId());
////SF//                                             assert !spokenForByKeyId.containsValue(abacc.getSpokenForUrn());
////SF//                                             assert !speakersKeyIds.contains(abacc.getTailKeyId());
////SF//                                             spokenForByKeyId.put(abacc.getHeadKeyId(), abacc.getSpokenForUrn());
////SF//                                             speakersKeyIds.add(abacc.getTailKeyId());
////SF//                                         }
////SF//                                         LOG.debug("Speaksfor chain search: spokenForByKeyId="+spokenForByKeyId);
////SF//                                         LOG.debug("Speaksfor chain search: speakersKeyIds="+speakersKeyIds);
////SF//                                         for (String speakersKeyId : speakersKeyIds)
////SF//                                             spokenForByKeyId.remove(speakersKeyId);
////SF//                                         LOG.debug("Speaksfor chain search: spokenForByKeyId="+spokenForByKeyId);
////SF//                                         if (spokenForByKeyId.size() == 1) {
////SF//                                             speakingForUser = spokenForByKeyId.values().iterator().next();
////SF//                                         } else {
////SF//                                             if (spokenForByKeyId.size() == 0) {
////SF//                                                 LOG.warn("Speaksfor chain search got 0 leftover speaking_for urns");
////SF//                                                 speakingForUser = null;
////SF//                                             } else {
////SF//                                                 //ignore that there are multiple
////SF//                                                 LOG.warn("Speaksfor chain search got "+spokenForByKeyId.size()+" leftover speaking_for urns");
////SF//                                                 speakingForUser = spokenForByKeyId.values().iterator().next();
////SF//                                             }
////SF//                                         }
////SF//                                         LOG.debug("Speaksfor chain search: speakingForUser="+speakingForUser);
////SF//                                     } else {
////SF//                                         speakingForUser = c.getSpokenForUrn();
////SF//                                     }
////SF//                                 }
////SF//                             }
////SF//                             if (speaksForCredential instanceof SfaCredential) {
////SF//                                 if (speaksForArg.length() > 1) {
////SF//                                     LOG.error("Multiple speaksfor credentials are not supported for type SfaCredential \"" + speaksForArg + "\"");
////SF//                                     err.println("Multiple speaksfor credentials are not supported for type SfaCredential \"" + speaksForArg + "\"");
////SF//                                     return 3;
////SF//                                 }
////SF//                                 supportedCredType = true;
////SF//                                 SfaCredential c = (SfaCredential) speaksForCredential;
////SF//                                 if (c.getTargetUrn() != null)
////SF//                                     currentSpeakingForUser = new GeniUrn(c.getTargetUrn());
////SF//                                 else {
////SF//                                     if (c.getTargetGid() != null) {
////SF//                                         Gid gid = new Gid(c.getTargetGid());
////SF//                                         if (gid.getSubjectUrn() != null) {
////SF//                                             currentSpeakingForUser = new GeniUrn(gid.getSubjectUrn());
////SF//                                         }
////SF//                                     }
////SF//                                 }
////SF//                                 speakingForUser = currentSpeakingForUser;
////SF//                             }
////SF//                             if (!supportedCredType) {
////SF//                                 LOG.error("Unsupported speaksfor credential type for \"" + speaksForArg + "\": "+speaksForCredential.getType());
////SF//                                 err.println("Unsupported speaksfor credential type for \"" + speaksForArg + "\": "+speaksForCredential.getType());
////SF//                                 return 3;
////SF//                             }
////SF//                         } catch (GeniUrn.GeniUrnParseException e) {
////SF//                             LOG.error("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\"", e);
////SF//                             err.println("Error parsing spoken for user URN from speaksFor credential from file \"" + speaksForArg + "\": " + e.getMessage());
////SF//                             return 3;
////SF//                         }
////SF// //                if (!speaksForCredential.validate()) {
////SF// //                    LOG.warn("WARNING: Provided speaksfor credential does not validate");
////SF// //                    err.println("WARNING: Provided speaksfor credential does not validate");
////SF// //                }
////SF//
////SF//                         boolean check;
////SF//                         try {
////SF//                             check = speaksForCredential.check(testbedInfoSource);
////SF//                         } catch (CredentialException e) {
////SF//                             LOG.warn("Error trying to check credential signature: ", e);
////SF//                             check = false;
////SF//                         }
////SF//                         if (!check) {
////SF//                             LOG.warn("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
////SF//                             err.println("WARNING: Provided speaksfor credential signature check failed. (frequently caused by: you modified the credential)");
////SF//                         }
////SF//
////SF//                         if (currentSpeakingForUser == null) {
////SF//                             LOG.warn("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
////SF//                             err.println("ERROR: could not determine the spoken for user in the speaksFor credential from file \"" + speaksForArg + "\". Cannot continue.");
////SF//                             return 3;
////SF//                         }
////SF//                         if (speakingForUser == null) {
////SF//                             LOG.warn("ERROR: could not determine the final spoken for user. Cannot continue.");
////SF//                             err.println("ERROR: could not determine the final spoken for user. Cannot continue.");
////SF//                             return 3;
////SF//                         }
////SF//                         if (!cli.silent)
////SF//                             out.println("Using speaksFor credential for user \"" + currentSpeakingForUser + "\"");
////SF//                         speaksForCredentials.add(speaksForCredential);
////SF//                     }
////SF//                 }
////SF//
////SF//                 if (!cli.silent && cli.speaksForArgs.length > 1)
////SF//                     out.println("You are using multiple speaksfor credentials. The final user being spoken for is \"" + speakingForUser + "\"");
////SF//             }
//
//            JFedConnectionProvider conPool = injector.getInstance(JFedConnectionProvider.class);
//            AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory
//                    automaticUserAndSliceApiWrapperFactory =
//                    injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class);
//            AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
//                    automaticAggregateManagerWrapperFactory =
//                    injector.getInstance(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class);
//
//            final UserAndSliceApiWrapper sa = automaticUserAndSliceApiWrapperFactory.create();
//            GeniUrn targetUserUrn = user.getUserUrn();
//            sliceInfoHT.put("userUrn", targetUserUrn);
//
////SF//            if (usesSpeaksFor) {
////SF//                sa.setSpeaksFor(speaksForCredentials, speakingForUser);
////SF//                targetUserUrn = speakingForUser;
////SF//                if (speaksForCredentials.size() == 1)
////SF//                    sliceInfoHT.put("speaksForCredential", speaksForCredentials.get(0));
////SF//                else {
////SF//                    int i = 1;
////SF//                    for (AnyCredential speaksForCredential : speaksForCredentials)
////SF//                        sliceInfoHT.put("speaksForCredential" + (i++), speaksForCredential);
////SF//                }
////SF//                sliceInfoHT.put("speakingForUserUrn", speakingForUser);
////SF//            }
//            String targetUserUrnString = targetUserUrn.toString();
//
//            Set<GeniUrn> shareUserUrns;
//            boolean shareSliceWithProjectUsers = false;
//            if (cli.shareSliceUsersArg != null) {
//                String tla = targetUserUrn.getEncodedTopLevelAuthority();
//                shareUserUrns = new HashSet<>();
//                String[] users = cli.shareSliceUsersArg.split(",");
//                for (String opt : users) {
//                    boolean validOpt = false;
//
//                    if (Objects.equals(opt, "PROJECTUSERS")) {
//                        if (!cli.silent)
//                            out.println("Will share slice with all users in the slice project.");
//                        shareSliceWithProjectUsers = true;
//                        validOpt = true;
//                    }
//                    if (!validOpt && GeniUrn.valid(opt)) {
//                        GeniUrn userUrn = GeniUrn.parse(opt);
//                        assert opt != null;
//                        shareUserUrns.add(userUrn);
//                        validOpt = true;
//                    }
//                    if (!validOpt && opt.matches("[a-zA-Z][a-zA-Z0-9]*") && opt.length() <= 8) {
//                        GeniUrn userUrn = GeniUrn.createGeniUrnFromEncodedParts(tla, "user", opt);
//                        if (!cli.silent)
//                            out.println("Assuming specified user \""+opt+"\" is: "+userUrn);
//                        shareUserUrns.add(userUrn);
//                        validOpt = true;
//                    }
//
//                    if (!validOpt) {
//                        err.println("The share-slice option you have specified (" + opt + ") is not valid a valid URN or username");
//                        cli.commandHelp();
//                        return 3;
//                    }
//                }
//            } else {
//                shareUserUrns = Collections.emptySet();
//            }
//            assert shareUserUrns != null;
//
//            List<AnyCredential> userCreds;
//            try {
//                userCreds = sa.getLocalUserCredentials(logger);
//                sliceInfoHT.put("userCredentials", userCreds);
//            } catch (JFedException e) {
//                return callFailed("Error retrieving user credential", e, err, callLogOutput);
//            }
//            if (cli.debug)
//                out.println(userCreds.size() + " user credentials retrieved.");
//            List<GeniUrn> existingSlices;
//            List<String> existingSlicesCompareStrings;
//            try {
//                existingSlices = sa.getSlicesForUser(logger, targetUserUrn);
//            } catch (JFedException e) {
//                return callFailed("Error retrieving slices for user", e, err, callLogOutput);
//            }
//            if (cli.debug)
//                out.println("User " + targetUserUrnString + " has the following slices: " + existingSlices);
//            if (existingSlices != null) {
//                existingSlicesCompareStrings = new ArrayList<>();
//                for (GeniUrn sliceUrn : existingSlices) {
//                    GeniUrn lowercaseSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
//                            sliceUrn.getEncodedTopLevelAuthority().toLowerCase(),
//                            sliceUrn.getEncodedResourceType(),
//                            sliceUrn.getEncodedResourceName().toLowerCase()
//                    );
//                    existingSlicesCompareStrings.add(lowercaseSliceUrn.toString());
//                }
//            } else
//                existingSlicesCompareStrings = null;
//
//            if (cli.isCommand(LIST_USER_INFO)) {
//                List<String> projectNames;
//                try {
//                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
//                } catch (JFedException e) {
//                    return callFailed("Error retrieving projects for user", e, err, callLogOutput);
//                }
//                if (cli.isJson) {
//                    Map<String,Object> userHt = new HashMap<>();
//                    userHt.put("urn", targetUserUrnString);
//                    userHt.put("slices", existingSlices);
//                    userHt.put("projects", projectNames);
//                    out.println(JSonHelper.jsonObjectToString(userHt));
//                }
//                if (cli.isTextOutput) {
//                    out.println("User Info: ");
//                    out.println("   Urn: "+targetUserUrnString);
//                    out.println("   Projects: "+ projectNames.stream().collect(Collectors.joining(",")));
//                    out.println("   Slices: "+ existingSlices.stream().map(GeniUrn::toString).collect(Collectors.joining(",")));
//                }
//
//                return 0;
//            }
//
//            List<String> projectNames = null;
//            if (Objects.equals(cli.projectNameArg, cli.CHOOSE_PROJECT_AUTOMATICALLY_OPTION)) {
//                try {
//                    projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
//
//                    if (cli.debug)
//                        out.println("You are a member of the following projects: " + projectNames);
//
//                    if (projectNames.isEmpty()) {
//                        err.println("WARNING: " + cli.CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but you do not seem to be a member of any project! Will use fallback: continue without project.");
//
//                        //fallback: no project
//                        cli.projectNameArg = null;
//                    } else {
//                        //select last project
//                        cli.projectNameArg = projectNames.get(projectNames.size() - 1);
//                        if (cli.debug || !cli.silent)
//                            out.println("Automatically selected project \"" + cli.projectNameArg + "\"");
//                    }
//                } catch (JFedException e) {
//                    //fallback: no project
//                    cli.projectNameArg = null;
//
//                    String message =
//                            "WARNING: " + cli.CHOOSE_PROJECT_AUTOMATICALLY_OPTION + " was set as project, but there was an error retrieving list of user projects. Will use fallback: continue without project.";
//                    if (e.getXmlRpcResult() != null)
//                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                    else {
//                        err.println(message + ":");
//                        e.printStackTrace(err);
//                    }
//                }
//            }
//
//            GeniUrn initialSliceUrn = cli.sliceIdArg.startsWith("urn:") ?
//                    GeniUrn.parse(cli.sliceIdArg) :
//                    GeniUrn.createGeniUrnFromEncodedParts(
//                            userAuth.getUrnTld() + (cli.projectNameArg == null ? "" : ":" + cli.projectNameArg),
//                            "slice",
//                            cli.sliceIdArg
//                    );
//
//            GeniUrn lowercaseInitialSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
//                    initialSliceUrn.getEncodedTopLevelAuthority().toLowerCase(),
//                    initialSliceUrn.getEncodedResourceType(),
//                    initialSliceUrn.getEncodedResourceName().toLowerCase()
//            );
////            final boolean sliceExists = existingSlices.contains(initialSliceUrn);
//            final boolean sliceExists = existingSlicesCompareStrings.contains(lowercaseInitialSliceUrn.toString());
//
//            if (!cli.silent && !(cli.isCommand(SHOW_SLICE_INFO) && sliceExists)&& !(cli.isCommand(CREATE_SLICE_ONLY) && !sliceExists))
//                out.println("slice " + initialSliceUrn + " " + (sliceExists ? "already exists" : "does not yet exist"));
//
//            //check if project exists
//            if ((cli.isAction(CREATE, CREATE_SLICE_ONLY)) && !sliceExists && initialSliceUrn.getEncodedSubAuthName() != null) {
//                try {
//                    if (projectNames == null)
//                        projectNames = sa.getSubAuthorityNames(logger, targetUserUrn);
//                    assert projectNames != null;
//
//                    List<String> projectNamesLowerCase = new ArrayList<>(projectNames.size()+1);
//                    for (String projectName : projectNames)
//                        projectNamesLowerCase.add(projectName.toLowerCase());
//
//                    if (!projectNamesLowerCase.contains(initialSliceUrn.getEncodedSubAuthName().toLowerCase()))
//                        err.println("WARNING: You do not seem to be a member if the project (\"" + initialSliceUrn.getEncodedSubAuthName() + "\") " +
//                                "in which you wish to create a slice. Projects you are member of: " + projectNames);
//                    else {
//                        if (!projectNames.contains(initialSliceUrn.getEncodedSubAuthName())) {
//                            String userProvidedProjectName = initialSliceUrn.getEncodedSubAuthName();
//                            for (String correctCaseProjectName : projectNames) {
//                                if (userProvidedProjectName.equalsIgnoreCase(correctCaseProjectName) && !Objects.equals(userProvidedProjectName, correctCaseProjectName)) {
//                                    GeniUrn orig = initialSliceUrn;
//                                    initialSliceUrn = GeniUrn.createGeniUrnFromEncodedParts(
//                                            initialSliceUrn.getEncodedTopLevelAuthority_withoutSubAuth()+":"+correctCaseProjectName,
//                                            initialSliceUrn.getEncodedResourceType(),
//                                            initialSliceUrn.getEncodedResourceName());
//                                    LOG.info("(CaseBugs): You are a member of \"" + userProvidedProjectName + "\", " +
//                                            "but that project is known with a different case: \"" + correctCaseProjectName + "\"." +
//                                            "jFed will change the case to avoid any SA/AM bugs. (though it shouldn't matter.). " +
//                                            "Changed slice URN from " + orig + " to " +initialSliceUrn);
//                                }
//                            }
//                        }
//
//                    }
//
//                    if (cli.debug)
//                        out.println("You are a member of the following projects: " + projectNames);
//                } catch (JFedException e) {
//                    String message = "Error retrieving list of user projects. (this error will be ignored)";
//                    if (e.getXmlRpcResult() != null)
//                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                    else {
//                        err.println(message + ":");
//                        e.printStackTrace(err);
//                    }
//                }
//            }
//
//            final GeniUrn sliceUrn = initialSliceUrn;
//
//            if (!(cli.isAction(CREATE, CREATE_SLICE_ONLY)) && !sliceExists) {
//                err.println("Error: The slice " + sliceUrn + " does not exist. Cannot continue.");
//                return 2;
//            }
//
//            if (cli.fake) {
//                out.println("Argument processing done. --cli.fake is set, so no further steps are taken.");
//                return 0;
//            }
//
//            List<AnyCredential> sliceCredentials;
//            Date sliceExpirationDate = null;
//            if (sliceExists) {
//                sliceInfoHT.put("sliceUrn", sliceUrn);
//                try {
//                    sliceCredentials = sa.getSliceCredentials(logger, sliceUrn);
//                    if (!sliceCredentials.isEmpty()) sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
//                    LOG.debug("ExperimenterCli requested slice credentials, and got "+sliceCredentials.size()+" credentials");
//                } catch (JFedException e) {
//                    return callFailed("Error retrieving slice credential", e, err, callLogOutput);
//                }
//
//                if (cli.rspecArg == null && !(cli.isAction(SHOW_SLICE_INFO, CREATE_SLICE_ONLY))) {
//                    //no RSpec provided, so get rspecComponentManagers from SLIVER_INFO (or equivalent)
//                    try {
//                        rspecComponentManagers = sa.getAggregatesForSlice(logger, sliceCredentials, sliceUrn).stream()
//                                .map(GeniUrn::getValue).collect(Collectors.toList());
//                    } catch (JFedException e) {
//                        return callFailed("Error retrieving slice credential", e, err, callLogOutput);
//                    }
//                    if (rspecComponentManagers.size() == 0) {
//                        err.println("WARNING: No RSpec was provided using the '--rspec' option, and the SA did not know of any AM's that have resources. " +
//                                "Few commands can now do anything. Will try anyway.");
//                    }
//                }
//
//                if (cli.isAction(RENEW, CREATE)) {
//                    Date currentSliceExpireDate = null;
//                    for (AnyCredential credential : sliceCredentials)
//                        if (credential.getExpiresDate() != null)
//                            currentSliceExpireDate = credential.getExpiresDate();
//                    if (currentSliceExpireDate != null && isBefore(currentSliceExpireDate, cli.requestedExpirationDate)) {
//                        try {
//                            if (cli.debug)
//                                out.println("Need to Renew slice, because the slice expires at " + RFC3339Util.dateToRFC3339String(currentSliceExpireDate, true, true, true) +
//                                        " while requested expiration date is " + RFC3339Util.dateToRFC3339String(cli.requestedExpirationDate, true, true, true));
//                            if (!cli.silent)
//                                out.println("Renewing slice, requested new expire date is "+RFC3339Util.dateToRFC3339String(cli.requestedExpirationDate, true, true, true)+" ...");
//                            List<AnyCredential> newSliceCredentials =
//                                    sa.renewSlice(logger, sliceCredentials, cli.requestedExpirationDate);
//                            sliceExpirationDate = cli.requestedExpirationDate; //fallback
//                            if (newSliceCredentials != null) {
//                                sliceCredentials = newSliceCredentials;
//                                if (cli.debug)
//                                    out.println("ExperimenterCli renewed the slice and got "+sliceCredentials.size()+" credentials");
//                                if (!sliceCredentials.isEmpty()) {
//                                    sliceExpirationDate = sliceCredentials.get(0).getExpiresDate();
//                                    if (sliceExpirationDate != null && isBefore(sliceExpirationDate, cli.requestedExpirationDate)) {
//                                        if (!cli.silent) {
//                                            err.println("WARNING: The slice was renewed, but the requested expiration date was not set. " +
//                                                    "Actual new expire date: "+RFC3339Util.dateToRFC3339String(sliceExpirationDate, true, true, true));
//                                        }
//                                    }
//                                }
//                            }
//                            else
//                                out.println("WARNING: did not receive renewed credential");
//                        } catch (JFedException e) {
//                            return callFailed("Error renewing slice credential", e, err, callLogOutput);
//                        }
//                    } else {
//                        if (currentSliceExpireDate != null) {
//                            if (cli.isCommand(RENEW))
//                                if (cli.debug)
//                                    out.println("No need to renew slice. It already expires at "+RFC3339Util.dateToRFC3339String(currentSliceExpireDate, true, true, true));
//                        } else {
//                            if (!cli.silent) {
//                                if (cli.isCommand(RENEW))
//                                    err.println("WARNING: expire date of slice is not known. Will not attempt renew!");
//                                else
//                                    err.println("WARNING: actual expire date of created slice is not known.");
//                            }
//                        }
//                    }
//                }
//
//                sliceInfoHT.put("sliceCredentials", sliceCredentials);
//
//                if (cli.isAction(SHOW_SLICE_INFO, CREATE_SLICE_ONLY)) {
//                    try {
//                        List<GeniUrn> sliceUsers = sa.getUsersForSlice(logger, sliceUrn, sliceCredentials);
//                        sliceInfoHT.put("sliceUsers", sliceUsers);
//                    } catch (JFedException e) {
//                        LOG.error("Error getting users for slice", e);
//                        err.println("Error getting users for slice, will ignore.");
//                    }
//
//                    try {
//                        List<GeniUrn> sliceAggs = sa.getAggregatesForSlice(logger, sliceCredentials, sliceUrn);
//                        sliceInfoHT.put("sliceAggregates", sliceAggs);
//                    } catch (JFedException e) {
//                        LOG.error("Error getting aggregates for slice", e);
//                        err.println("Error getting aggregates for slice, will ignore.");
//                    }
//
//                    try {
//                        Date sliceFoundExpirationDate = sa.getSliceExpiration(logger, sliceCredentials, sliceUrn);
//                        if (sliceFoundExpirationDate != null)
//                            sliceInfoHT.put("sliceExpiration", RFC3339Util.dateToRFC3339String(sliceFoundExpirationDate));
//                    } catch (JFedException e) {
//                        LOG.error("Error getting expiration date for slice", e);
//                        err.println("Error getting expiration date, will ignore.");
//                    }
//                }
//            } else {
//                assert cli.isAction(CREATE, CREATE_SLICE_ONLY);
//                if (!cli.createSliceArg && cli.isCommand(CREATE)) {
//                    err.println("Error: The slice " + sliceUrn + " does not exist, and you have not specified the --create-slice option.");
//                    return 2;
//                }
//                UserAndSliceApiWrapper.SliceInfo sliceInfo;
//                try {
//                    sliceInfo =
//                            sa.createSlice(logger, sliceUrn.getEncodedResourceName(), cli.requestedExpirationDate, cli.projectNameArg);
//                    if (!sliceInfo.getCredentials().isEmpty()) {
//                        sliceExpirationDate = sliceInfo.getCredentials().get(0).getExpiresDate();
//                    }
//                } catch (JFedException e) {
//                    return callFailed("Error creating slice", e, err, callLogOutput);
//                }
//                sliceCredentials = sliceInfo.getCredentials();
//                LOG.debug("ExperimenterCli created a slice, and got "+sliceCredentials.size()+" credentials");
//
//                sliceInfoHT.put("sliceCredentials", sliceCredentials);
//                sliceInfoHT.put("sliceUrn", sliceInfo.getUrn());
//                sliceInfoHT.put("sliceExpiration", cli.requestedExpirationDate);
//            }
//            if (sliceCredentials == null || sliceCredentials.isEmpty()) {
//                err.println("Error: Did not receive slice credentials. (but call was successful)");
//                return 2;
//            }
//            LOG.debug("ExperimenterCLI has "+sliceCredentials.size()+" slice credentials");
//
//
//
//            if (shareSliceWithProjectUsers) {
//                if (cli.projectNameArg == null) {
//                    if (!cli.silent) {
//                        out.println("NOTE: Slice will NOT be shared with all members of project, because no project is specified.");
//                    }
//                } else {
//                    try {
//                        List<GeniUrn> projectUsers = sa.getUsersForSubAuthority(logger, cli.projectNameArg);
//                        shareUserUrns.addAll(projectUsers);
//                        if (!cli.silent) {
//                            out.println("Slice will be shared with all members of project \"" + cli.projectNameArg + "\": " + projectUsers);
//                        }
//                    } catch (UnsupportedOperationException e) {
//                        err.println("Cannot get project members of \"" + cli.projectNameArg + "\" to share slice with. -> Will not share slice with project members. " +
//                                "Reason: API does not support retrieving the project members.");
//                    } catch (JFedException e) {
//                        LOG.error("Failed to get project members of \"" + cli.projectNameArg + "\" to share slice with. -> Will not share slice with project members.", e);
//                        err.println("Failed to get project members of \"" + cli.projectNameArg + "\" to share slice with. Will not share slice with project members. Exception: " + e.getMessage());
//                        e.printStackTrace(err);
//                    }
//                }
//            }
//            if (cli.debug) {
//                out.println("Slice will be shared with: "+shareUserUrns);
//            }
//
//            if (!shareUserUrns.isEmpty()) {
//                if (!cli.silent)
//                    out.println("Will share existing slice with: "+shareUserUrns);
//                for (GeniUrn userUrn : shareUserUrns) {
//                    try {
//                        sa.shareSlice(logger, sliceUrn, sliceCredentials, userUrn);
//                    } catch (JFedException e) {
//                        LOG.error("Failed to share slice with \""+userUrn+"\"", e);
//                        err.println("Failed to share slice with \"" + userUrn + "\": " + e.getMessage());
//                        e.printStackTrace(err);
//                    }
//                }
//            }
//
//
//            GeniUrn newSliceUrn = null;
//            //see FEDIBBTDEV-2071: change slice URN to match slice URN in credential
//            for (AnyCredential cred : sliceCredentials) {
//                if (cred instanceof SfaCredential) {
//                    SfaCredential c = (SfaCredential) cred;
//                    if (c.getTargetUrn() != null && c.getTargetUrn().equalsIgnoreCase(sliceUrn.toString()) && !Objects.equals(c.getTargetUrn(), sliceUrn.toString())) {
//                        String message = "(CaseBugs): slice URN ("+sliceUrn+") does not match URN in credential ("+c.getTargetUrn()+"). " +
//                                "Will automatically fix this by changing slice URN. " +
//                                "(Otherwise, some AM's might give strange errors)";
////                        if (!cli.silent)
//                        if (cli.debug)
//                            err.println("WARNING: "+message);
//                        LOG.warn(message);
//
//                        newSliceUrn = GeniUrn.parse(c.getTargetUrn());
//                        if (newSliceUrn != null) {
//                            break;
//                        }
//                        else
//                            err.println("ERROR: slice URN in credential is not a valid URN. This should not happen.");
//                    }
//                }
//                if (cred instanceof AbacCredential) {
//                    AbacCredential c = (AbacCredential) cred;
//                    //cannot do this for ABAC
//                }
//            }
//            final GeniUrn usedSliceUrn = newSliceUrn == null ? sliceUrn : newSliceUrn;
//            sliceInfoHT.put("sliceUrn", usedSliceUrn);
//
//            if (cli.isAction(CREATE_SLICE_ONLY, SHOW_SLICE_INFO)) {
//                if (cli.isCommand(CREATE_SLICE_ONLY) && sliceExists) {
//                    err.println("Error: Slice already exists.");
//                }
//
//                if (cli.isJson) {
//                    String jsonOutput = XmlRpcPrintUtil.xmlRpcObjectToString(sliceInfoHT);
//                    out.println(jsonOutput);
//                } else {
//                    assert cli.isTextOutput;
//                    out.println("User URN: "+sliceInfoHT.get("userUrn"));
////                    out.println("User Credentials: "+sliceInfoHT.get("userCredentials"));
//                    out.println("Slice URN: "+sliceInfoHT.get("sliceUrn"));
//                    out.println("Slice URN: "+sliceInfoHT.get("sliceExpiration"));
//                    List sliceCreds = (List) sliceInfoHT.get("sliceCredentials");
//                    int i = 0;
//                    if (sliceCreds != null)
//                        for (Object c : sliceCreds) {
//                            if (c != null && c instanceof AnyCredential) {
//                                AnyCredential ac = (AnyCredential) c;
//                                out.println("Slice Credential " + (++i) + ":\n"+ac.getCredentialXml()+"\n\nEND SLICE CREDENTIAL\n");
//                            }
//                        }
//                }
//
//                if (cli.isCommand(CREATE_SLICE_ONLY) && sliceExists)
//                    return 1;
//                return 0;
//            }
//
//            if (cli.isCommand(CREATE) && cli.sliceRecoverInfoArg != null) {
//                if (cli.debug)
//                    out.println("Writing slice recovery info file \"" + cli.sliceRecoverInfoArg + "\"");
//
//                try {
//                    FileConfiguration conf = new PropertiesConfiguration();
//                    conf.setEncoding("UTF-8");
//                    conf.setProperty(KEY_EXPIRATION_TIME,
//                            sliceExpirationDate == null ? "unknown"
//                                    : ISO_OFFSET_DATE_TIME.format(ZonedDateTime.ofInstant(sliceExpirationDate.toInstant(), ZoneId.systemDefault())));
//                    conf.setProperty(KEY_REQUEST_RSPEC, rspec.getXmlString());
//                    conf.save(new File(cli.sliceRecoverInfoArg));
//                } catch (ConfigurationException e) {
//                    LOG.error("Error writing slice recovery info to \"" + cli.sliceRecoverInfoArg + "\"", e);
//                    err.println("ERROR (non fatal): Exception writing slice recovery info to \"" + cli.sliceRecoverInfoArg + "\": " + e.getMessage());
//                }
//            }
//
//
//            if (cli.fakeSliver) {
//                out.println("Argument and slice processing done. --cli.fake-sliver is set, so no further steps are taken.");
//                return 0;
//            }
//
//            final List<ExperimenterCli.AutoShareLanInfo> lanToShare = new ArrayList<>();
//            if (cli.isCommand(CREATE)) {
//                try {
//                    modelRspec = getModelRspec(modelRspec, rspec);
//
//                    for(RspecLink link : modelRspec.getLinks()) {
//                        if (link.hasAutoShareLan()) {
//                            for (String shareName : link.getAutoShareLanNames()) {
//                                String clientId = link.getClientId();
//                                if (clientId == null) {
//                                    LOG.warn("Cannot share link that has no client_id");
//                                } else {
//                                    if (link.getComponentManagerUrns().size() > 1) {
//                                        LOG.warn("Cannot share link ("+clientId+") that has multiple component_managers");
//                                    } else {
//                                        if (link.getComponentManagerUrns().isEmpty()) {
//                                            LOG.warn("Cannot share link ("+clientId+") that has no component_managers");
//                                        } else {
//                                            GeniUrn cm = link.getComponentManagerUrns().get(0);
//                                            ExperimenterCli.AutoShareLanInfo asli = new ExperimenterCli.AutoShareLanInfo(cm, clientId, shareName);
//                                            lanToShare.add(asli);
//                                        }
//                                    }
//                                }
//                            }
//                        }
//                    }
//                } catch (RspecParseException e) {
//                    LOG.error("Problem parsing RSpec to automatically share LAN. Will not automatically share LAN.", e);
//                    lanToShare.clear();
//                }
//            }
//
//            final List<ExperimenterCli.AnsiblePlaybookInfo> ansiblePlaybooksToExecute = new ArrayList<>();
//            if ((cli.isAction(CREATE, MANIFEST)) && !cli.noPlaybookToExecute) {
//                if (cli.ansibleExtraPlaybooks != null) {
//                    LOG.debug("Processing "+cli.ansibleExtraPlaybooks.length+" command line extra playbooks");
//                    for (String ansibleExtraPlaybook : cli.ansibleExtraPlaybooks) {
//                        String[] parts = ansibleExtraPlaybook.split(" ", 2);
//                        String source = parts[0];
//                        String outputFilename = parts.length > 1 ? parts[1] : null;
//                        ExecuteAnsiblePlaybook executeAnsiblePlaybook = new BasicExecuteAnsiblePlaybook(source, outputFilename);
//                        ExperimenterCli.AnsiblePlaybookInfo info = new ExperimenterCli.AnsiblePlaybookInfo(executeAnsiblePlaybook);
//                        try {
//                            info.loadContent(cli.ansibleAllowPlaybookInputfile);
//                            info.checkFilename(false /*mustBeBasename*/, cli.ansibleAllowAnyPlaybookOutputFile);
//                        } catch (ExperimenterCli.AnsiblePlaybookInfo.AnsiblePlaybookException e) {
//                            LOG.debug("Error processing ExecuteAnsiblePlaybook", e);
//                            err.println("Error processing --ansible-add-playbook: " + e.getMessage());
//                            return 2;
//                        }
//                        ansiblePlaybooksToExecute.add(info);
//                    }
//                }
//
//                if (rspec != null && cli.ansibleExecuteRspecPlaybooks) {
//                    try {
//                        modelRspec = getModelRspec(modelRspec, rspec);
//
//                        for (ExecuteAnsiblePlaybook executeAnsiblePlaybook : modelRspec.getExecuteAnsiblePlaybooks()) {
//                            ExperimenterCli.AnsiblePlaybookInfo info = new ExperimenterCli.AnsiblePlaybookInfo(executeAnsiblePlaybook);
//                            try {
//                                info.loadContent(cli.ansibleAllowPlaybookInputfile);
//                                info.checkFilename(false /*mustBeBasename*/, cli.ansibleAllowAnyPlaybookOutputFile);
//                            } catch (ExperimenterCli.AnsiblePlaybookInfo.AnsiblePlaybookException e) {
//                                LOG.debug("Error processing ExecuteAnsiblePlaybook", e);
//                                err.println("Error processing " + RspecXmlConstants.Q_JFED_EXECUTE_ANSIBLE_PLAYBOOK.getLocalPart() + ": " + e.getMessage());
//                                return 2;
//                            }
//                            ansiblePlaybooksToExecute.add(info);
//                        }
//                    } catch (RspecParseException e) {
//                        LOG.error("Problem parsing RSpec to automatically execute ansible playbooks. Will not automatically execute ansible playbooks.", e);
//                        ansiblePlaybooksToExecute.clear();
//                    }
//                }
//
//                Set<String> sources = new HashSet<>();
//                for (Iterator<ExperimenterCli.AnsiblePlaybookInfo> it = ansiblePlaybooksToExecute.iterator(); it.hasNext();) {
//                    ExperimenterCli.AnsiblePlaybookInfo i = it.next();
//                    if (!sources.add(i.getExecuteAnsiblePlaybook().getSource())) {
//                        err.println("Warning: duplicate playbook will be ignored: " + i.getExecuteAnsiblePlaybook());
//                        it.remove();
//                    }
//                }
//
//                if (cli.debug) {
//                    out.println("There are "+ansiblePlaybooksToExecute.size()+" ansible playbooks to execute.");
//                }
//            }
//
//
//            List<UserSpec> userSpecs = new ArrayList<>();
//            if (cli.isCommand(CREATE)) {
//                if (cli.sshKeysUserServerKeys) {
//                    try {
//                        List<String> userKeys = sa.getSshKeysForUser(logger, targetUserUrn);
//                        addToUserSpecs(userSpecs, targetUserUrn, userKeys);
//                        if (cli.debug)
//                            out.println("Added " + userKeys.size() + " keys from server.");
//                    } catch (JFedException e) {
//                        return callFailed("Error retrieving ssh keys for user " + targetUserUrn, e, err, callLogOutput);
//                    }
//                }
//                if (!shareUserUrns.isEmpty() && cli.sshKeysSharedUserKeys) {
//                    for (GeniUrn shareUserUrn : shareUserUrns) {
//                        try {
//                            List<String> userKeys = sa.getSshKeysForUser(logger, shareUserUrn);
//                            if (userKeys != null && !userKeys.isEmpty()) {
//                                addToUserSpecs(userSpecs, shareUserUrn, userKeys);
//                                if (cli.debug)
//                                    out.println("Added " + userKeys.size() + " keys from server for user \"" + shareUserUrn + "\".");
//                            } else {
//                                if (!cli.silent)
//                                    err.println("WARNING: No SSH keys found for user \"" + shareUserUrn + "\".");
//                            }
//                        } catch (JFedException e) {
//                            LOG.error("Error retrieving ssh keys for user \"" + shareUserUrn+"\"", e);
//                            err.println("Error retrieving ssh keys for user \"" + shareUserUrn+"\": "+e.getMessage());
//                            e.printStackTrace(err);
//                        }
//                    }
//                }
//                if (cli.sshKeysUserCert) {
//                    if (!usesSpeaksFor) {
//                        String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey());
//                        addToUserSpecs(userSpecs, targetUserUrn, userKey);
//                        if (cli.debug)
//                            out.println("Adding userspec from user certificate.");
//                    } else {
//                        PublicKey pubKey = null;
//                        for (AnyCredential speaksForCredential : speaksForCredentials) {
//                            assert speaksForCredential != null;
//                            PublicKey curPubKey = null;
//                            GeniUrn currentSpeakingForUser = null;
//                            if (speaksForCredential instanceof AbacCredential) {
//                                AbacCredential c = (AbacCredential) speaksForCredential;
//                                curPubKey = c.getSpokenForPubKey();
//
//                                if (c.getSpokenForUrn() != null)
//                                    currentSpeakingForUser = c.getSpokenForUrn();
//                            }
//                            try {
//                                if (speaksForCredential instanceof SfaCredential) {
//                                    SfaCredential c = (SfaCredential) speaksForCredential;
//                                    if (c.getTargetUrn() != null)
//                                        currentSpeakingForUser = new GeniUrn(c.getTargetUrn());
//                                    if (c.getTargetGid() != null) {
//                                        Gid gid = new Gid(c.getTargetGid());
//                                        if (gid.getCertificate() != null)
//                                            curPubKey = gid.getCertificate().getPublicKey();
//                                        if (currentSpeakingForUser == null && gid.getSubjectUrn() != null) {
//                                            currentSpeakingForUser = new GeniUrn(gid.getSubjectUrn());
//                                        }
//                                    }
//                                }
//                            } catch (GeniUrn.GeniUrnParseException e) {
//                                LOG.error("Ignoring speaksfor cert target URN parse exception", e);
//                            }
//
//                            if (curPubKey != null && currentSpeakingForUser != null && Objects.equals(currentSpeakingForUser, speakingForUser)) {
//                                pubKey = curPubKey;
//                            }
//                        }
//                        if (pubKey != null) {
//                            String userKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(pubKey);
//                            addToUserSpecs(userSpecs, targetUserUrn, userKey);
//                            if (cli.debug)
//                                out.println("Adding userspec from speaksFor SfaCredential");
//                        } else {
//                            err.println("WARNING: Could not fetch user public key from user speaksFor certificate(s).");
//                        }
//                    }
//                }
//                if (cli.sshKeysRspec) {
//                    try {
//                        modelRspec = getModelRspec(modelRspec, rspec);
//                        List<UserSpec> extraUserSpec = modelRspec.getSshKeys();
//                        if (cli.debug)
//                            out.println("Adding " + extraUserSpec.size() + " userspecs from RSpec");
//                        for (UserSpec us : extraUserSpec) {
//                            GeniUrn userUrn = GeniUrn.parse(us.getUrn());
//                            addToUserSpecs(userSpecs, userUrn, us.getSshKey());
//                        }
//                    } catch (RspecParseException e) {
//                        err.println("WARNING: Error parsing specified RSpec. The users could not be extracted from this RSpec. Note: This tool will continue, and the RSpec will still be used to create slivers.");
//                    }
//                }
//                if (cli.debug) {
//                    out.println("The following user specification will be added to all create sliver calls:");
//                    for (UserSpec us : userSpecs) {
//                        out.println("  User " + us.getUrn() + " with " + us.getSshKey().size() + " keys");
//                        for (String key : us.getSshKey())
//                            out.println("     " + key);
//                    }
//                }
//            }
//
//            //create amWrappers
//            List<AggregateManagerWrapper> amWrappers = new ArrayList<>();
//            for (String rspecComponentManager : rspecComponentManagers) {
//                GeniUrn rspecComponentManagerUrn = GeniUrn.parse(rspecComponentManager);
//
//                Server currentAuthority =
//                        authorityFinder.findByUrn(rspecComponentManager, AuthorityFinder.Purpose.CREATE_SLIVER);
//                if (currentAuthority == null) {
//                    err.println("Tested authority URN is unknown: \"" + rspecComponentManager + "\"");
//                    return 2;
//                }
//                assert currentAuthority != null;
//
//                final AggregateManagerWrapper am;
//                if (cli.amApiVersionArg == null) {
//                    am = automaticAggregateManagerWrapperFactory.create(currentAuthority);
//                } else {
//                    if (Objects.equals(cli.amApiVersionArg, "2"))
//                        am = new AMv2Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
//                    else {
//                        assert Objects.equals(cli.amApiVersionArg, "3");
//                        am = new AMv3Wrapper(logger, userProvider, conPool, jFedPreferences, currentAuthority);
//                    }
//                }
//                if (usesSpeaksFor) {
//                    am.setSpeaksFor(speaksForCredentials, speakingForUser);
//                }
//                amWrappers.add(am);
//            }
//
//
//
//            if (cli.isCommand(CREATE) && cli.abortIfSliversExist) {
//                if (cli.debug) {
//                    out.println("--abort-if-slivers-exist option is set: will first check if slivers already exist");
//                }
//                for (AggregateManagerWrapper am : amWrappers) {
//                    Server currentAuthority = am.getAmServer();
//                    String rspecComponentManager = currentAuthority.getDefaultComponentManagerUrn();
//                    if (!cli.silent)
//                        out.println("Contacting  " + rspecComponentManager + "...");
//
//                    try {
//                        LOG.debug("ExperimenterCLI is calling status with " + sliceCredentials.size() + " credentials");
//                        StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);
//
//                        if (!statusDetails.getAllSliverUrns().isEmpty()) {
//                            err.println("Slivers exist at "+rspecComponentManager+" and --abort-if-slivers-exist is set: will not create any new slivers");
//                            return -1;
//                        }
//                        if (cli.debug) {
//                            out.println("No slivers at " + rspecComponentManager+"  (got empty sliver list)");
//                        }
//                    } catch (JFedException e) {
//                        //error is expected: normally means there are no slivers
//                        if (cli.debug) {
//                            out.println("No known slivers at " + rspecComponentManager + "  (got error)");
//                        }
//                    }
//                }
//            }
//
//            ManifestRspecMerger manifestRspecMerger = new ManifestRspecMerger(ModelRspecType.BASIC);
//
//            Map outerStatusHt = null;
//            Map outerStatusAmDetailHt = null;
//            Date earliestSliverExpireDate = null;
//            if (cli.isCommand(STATUS)) {
//                outerStatusHt = new HashMap();
//                outerStatusHt.put("sliceUrn", sliceUrn);
//                if (sliceExpirationDate != null)
//                    outerStatusHt.put("sliceExpireDate", RFC3339Util.dateToRFC3339String(sliceExpirationDate));
//                outerStatusAmDetailHt = new HashMap();
//                outerStatusHt.put("AMs", outerStatusAmDetailHt);
//            }
//
//            for (AggregateManagerWrapper am : amWrappers) {
//                Server currentAuthority = am.getAmServer();
//                String rspecComponentManager = currentAuthority.getDefaultComponentManagerUrn();
//                GeniUrn rspecComponentManagerUrn = currentAuthority.getDefaultComponentManagerAsGeniUrn();
//                if (!cli.silent)
//                    out.println("Contacting  " + rspecComponentManager + "...");
//
//                if (cli.isCommand(CREATE)) {
//                    try {
//                        String requestRspec = rspec.getXmlString();
//
//                        String manifest;
//                        if (am.hasSeperateAllocateAndProvision()) {
//                            AggregateManagerWrapper.AllocateReply allocateReply =
//                                    am.allocate(usedSliceUrn, sliceCredentials, requestRspec, cli.requestedExpirationDate);
//
//                            if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
//                                sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmServer().getDefaultComponentManagerAsGeniUrn(), allocateReply.getSliverUrns(), cli.requestedExpirationDate);
//
//                            manifest =
//                                    am.provisionAndStart(usedSliceUrn, sliceCredentials, userSpecs, cli.requestedExpirationDate);
//
//                            if (manifest != null) {
//                                Collection<GeniUrn> sliverUrns = ManifestSliverFinder.findSliversInManifest(manifest);
//                                if (allocateReply != null && allocateReply.getSliverUrns() != null && !allocateReply.getSliverUrns().isEmpty())
//                                    sliverUrns.removeAll(allocateReply.getSliverUrns());
//                                if (!sliverUrns.isEmpty())
//                                    sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmServer().getDefaultComponentManagerAsGeniUrn(), sliverUrns, cli.requestedExpirationDate);
//                            }
//                        } else {
//                            manifest =
//                                    am.createSliver(usedSliceUrn, sliceCredentials, requestRspec, userSpecs, cli.requestedExpirationDate);
//
//                            if (manifest != null) {
//                                Collection<GeniUrn> sliverUrns = ManifestSliverFinder.findSliversInManifest(manifest);
//                                sa.registerAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, am.getAmServer().getDefaultComponentManagerAsGeniUrn(), sliverUrns, cli.requestedExpirationDate);
//                            }
//                        }
//
//                        if (!cli.silent)
//                            out.println("Sliver at  " + rspecComponentManager + " is created and initializing...");
//
//                        manifestRspecMerger.setManifest(currentAuthority.getDefaultComponentManagerAsGeniUrn(), new ManifestRspecSource(manifest, ModelRspecType.BASIC));
//                    } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
//                        if (cli.deleteOnCreateFailure)
//                            deleteAll(amWrappers, cli.silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
//                        return callFailed("Not enough resources available to create sliver at " + currentAuthority.getDefaultComponentManagerUrn(), e, err, callLogOutput);
//                    } catch (JFedException e) {
//                        if (cli.deleteOnCreateFailure)
//                            deleteAll(amWrappers, cli.silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
//                        return callFailed("Error creating sliver at " + currentAuthority.getDefaultComponentManagerUrn(), e, err, callLogOutput);
//                    }
//                }
//
//                if (cli.isCommand(DELETE))
//                    try {
//                        am.deleteSliver(usedSliceUrn, sliceCredentials);
//
//                        if (!cli.silent)
//                            out.println("Slivers at  " + rspecComponentManager + " have been deleted.");
//
//
//                        try {
//                            sa.unregisterAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, rspecComponentManagerUrn, null/*sliverUrns*/);
//
//                            if (!cli.silent)
//                                out.println("Slivers at  " + rspecComponentManager + " has been unregistered.");
//                        } catch (JFedException e) {
//                            String message =
//                                    "Error unregistering sliver at " + rspecComponentManager + ". (this error will be ignored)";
//                            if (e.getXmlRpcResult() != null)
//                                err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                            else {
//                                err.println(message + ":");
//                                e.printStackTrace(err);
//                            }
//                        }
//                    } catch (JFedException e) {
//                        String message =
//                                "Error deleting sliver at " + rspecComponentManager + ". (this error will be ignored)";
//                        if (e.getXmlRpcResult() != null)
//                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                        else {
//                            err.println(message + ":");
//                            e.printStackTrace(err);
//                        }
//                    }
//
//                if (cli.isCommand(STATUS))
//                    try {
//                        LOG.debug("ExperimenterCLI is calling status with " + sliceCredentials.size() + " credentials");
//                        StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);
//
//                        if (!cli.silent && cli.isTextOutput)
//                            out.println("Sliver at  " + rspecComponentManager + " has status " + statusDetails.getGlobalStatus());
//
//                        if (cli.isJson || cli.isTextOutput) {
//                            Date earliestAMSliverExpireDate = statusDetails.getEarliestKnownExpires();
//                            if (earliestSliverExpireDate == null)
//                                earliestSliverExpireDate = earliestAMSliverExpireDate;
//                            else {
//                                if (earliestAMSliverExpireDate != null && earliestAMSliverExpireDate.before(earliestSliverExpireDate))
//                                    earliestSliverExpireDate = earliestAMSliverExpireDate;
//                            }
//                            Map outerAMHt = new HashMap();
//                            outerAMHt.put("amUrn", am.getAmServer().getDefaultComponentManagerUrn());
//                            outerAMHt.put("amGlobalSliverStatus", statusDetails.getGlobalStatus());
//                            if (earliestAMSliverExpireDate != null)
//                                outerAMHt.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestAMSliverExpireDate));
//                            List detailStatus = new ArrayList();
//                            for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetails.getStatusBySliverUrn().entrySet()) {
//                                Map detailHt = new HashMap();
//                                detailHt.put("urn", detail.getKey());
//                                detailHt.put("type", "sliver");
//                                detailHt.put("status", detail.getValue() + "");
//                                detailStatus.add(detailHt);
//                            }
//                            for (Map.Entry<GeniUrn, StatusDetails.SliverStatus> detail : statusDetails.getStatusByComponentUrn().entrySet()) {
//                                Map detailHt = new HashMap();
//                                detailHt.put("urn", detail.getKey());
//                                detailHt.put("type", "component");
//                                detailHt.put("status", detail.getValue() + "");
//                                detailStatus.add(detailHt);
//                            }
//                            outerAMHt.put("details", detailStatus);
//                            outerStatusAmDetailHt.put(am.getAmServer().getDefaultComponentManagerUrn(), outerAMHt);
//                        }
//                    } catch (JFedException e) {
//                        String message =
//                                "Error getting status of sliver at " + rspecComponentManager + ". (this error will be ignored)";
//                        if (e.getXmlRpcResult() != null) {
//                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
//                        } else {
//                            err.println(message + ":");
//                            e.printStackTrace(err);
//                        }
//                    }
//
//
//                if (cli.isCommand(MANIFEST))
//                    try {
//                        LOG.debug("ExperimenterCLI is fetching manifest with " + sliceCredentials.size() + " credentials");
//                        String manifestString = am.describe(usedSliceUrn, sliceCredentials);
//
//                        if (manifestString != null)
//                            manifestRspecMerger.setManifest(rspecComponentManagerUrn, new ManifestRspecSource(manifestString, ModelRspecType.BASIC));
//
//                        if (!cli.silent && manifestString != null)
//                            out.println("Manifest for " + rspecComponentManager + " received.");
//                        if (manifestString == null)
//                            err.println("Error fetching manifest for " + rspecComponentManager + "");
//                    } catch (JFedException e) {
//                        String message =
//                                "Error fetching manifest at " + rspecComponentManager + ". (this error will be ignored, but the final manifest might be missing information)";
//                        if (e.getXmlRpcResult() != null) {
//                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
//                        } else {
//                            err.println(message + ":");
//                            e.printStackTrace(err);
//                        }
//                    }
//
//                if (cli.isCommand(POA)) {
//                    if (rspecComponentManagerUrn == null ||
//                            !Objects.equals(rspecComponentManagerUrn.getTopLevelAuthority_withoutSubAuth(), cli.poaTargetSliverUrn.getTopLevelAuthority_withoutSubAuth())) {
//                        if (!cli.silent)
//                            out.println("No PerformOperationalAction command on  " + rspecComponentManager + ".");
//                        continue;
//                    }
//                    if (!cli.silent)
//                        out.println("Executing PerformOperationalAction command on " + rspecComponentManager + "...");
//                    try {
//                        if (cli.poaActionArg.equalsIgnoreCase("ReloadOS")) {
//                            if (!am.isReloadOSSupported()) {
//                                err.println("The AM at " + rspecComponentManager + " does not support " + cli.poaActionArg + ". Will try anyway, but the call will likely fail.");
//                            }
//                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
//                            am.reloadOS(cli.poaTargetSliverUrn, sliceCredentials);
//                        }
//                        if (cli.poaActionArg.equalsIgnoreCase("Restart")) {
//                            if (!am.isRestartSupported()) {
//                                err.println("The AM at " + rspecComponentManager + " does not support " + cli.poaActionArg + ". Will try anyway, but the call will likely fail.");
//                            }
//                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
//                            am.restart(cli.poaTargetSliverUrn, sliceCredentials);
//                        }
//                        if (cli.poaActionArg.equalsIgnoreCase("ConsoleUrl")) {
//                            if (!am.isOpenConsoleSupported()) {
//                                err.println("The AM at " + rspecComponentManager + " does not support " + cli.poaActionArg + ". Will try anyway, but the call will likely fail.");
//                            }
//                            LOG.debug("ExperimenterCLI is calling PerformOperationalAction with " + sliceCredentials.size() + " credentials");
//                            URL consoleURL = am.openConsole(cli.poaTargetSliverUrn, sliceCredentials);
//                            if (consoleURL != null)
//                                out.println("The console URN for " + cli.poaTargetSliverUrn + " is " + consoleURL.toExternalForm());
//                            else
//                                out.println("Error: No console URN was returned for " + cli.poaTargetSliverUrn + "");
//                        }
//                        if (!cli.silent)
//                            out.println("PerformOperationalAction call completed.");
//                    } catch (JFedException e) {
//                        String message =
//                                "Error in PerformOperationalAction " + cli.poaActionArg + " of " + cli.poaTargetSliverUrn + ".";
//                        if (e.getXmlRpcResult() != null) {
//                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                            err.println("Request was: " + "\"" + e.getXmlRpcResult().getRequestXmlRpcString() + "\"");
//                        } else {
//                            err.println(message + ":");
//                            e.printStackTrace(err);
//                        }
//                    }
//                }
//
//                if (cli.isCommand(RENEW)) {
//                    Date requestedSliverExpirationDate = sliceExpirationDate == null ? cli.requestedExpirationDate : sliceExpirationDate;
//                    assert requestedSliverExpirationDate != null;
//
//                    try {
//                        if (!cli.silent) {
//                            out.print("Renewing sliver at  " + rspecComponentManager + " to expire at " +
//                                    RFC3339Util.dateToRFC3339String(requestedSliverExpirationDate, true, true, true));
//                            if (sliceExpirationDate != null)
//                                out.println(" (Slice expired at " +
//                                        RFC3339Util.dateToRFC3339String(sliceExpirationDate, true, true, true) + ")");
//                            else
//                                out.println();
//                        }
//
//                        am.renewSlivers(usedSliceUrn, null/*no sliverUrns known*/, sliceCredentials, requestedSliverExpirationDate);
//
//                        if (!cli.silent)
//                            out.println("Sliver at  " + rspecComponentManager + " has been renewed successfully");
//
//                    } catch (JFedException e) {
//                        successExitStatus = false;
//                        String message =
//                                "Error renewing sliver at " + rspecComponentManager + ". (jFed will first continue Renewing the other AMs and fail afterwards)\n" +
//                                        "WARNING: Your slivers (= resources) might now expire before your slice itself expires. You can use the \"status\" command to check this.";
//                        if (e.getXmlRpcResult() != null)
//                            err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                        else {
//                            err.println(message + ":");
//                            e.printStackTrace(err);
//                        }
//                    }
//
//                    if (sa.isRegisterAggregatesForSliceSupported())
//                        try {
//                            StatusDetails statusDetails = am.status(usedSliceUrn, sliceCredentials);
//
//                            Collection<GeniUrn> sliverUrns = statusDetails.getAllSliverUrns();
//                            Date firstExpires = statusDetails.getEarliestKnownExpires();
//                            Date lastExpires = statusDetails.getLatestKnownExpires();
//                            Date updateExpireDate = requestedSliverExpirationDate;
//                            if (firstExpires != null && lastExpires != null) {
//                                updateExpireDate = lastExpires; //better register them too long than too short
//                                if (!Objects.equals(firstExpires, lastExpires)) {
//                                    if (!cli.silent)
//                                        err.println("WARNING: The slivers at "+rspecComponentManager+" do not expire at the same time. " +
//                                                "first expire time = "+firstExpires+" last expire time = "+lastExpires+"" +
//                                                " -> Will update all slivers as expiring at "+updateExpireDate);
//                                }
//                            }
//
//                            sa.updateSliverExpirationDateInfoForSlice(logger, sliceCredentials, usedSliceUrn, sliverUrns, updateExpireDate);
//                        } catch (Exception e) {
//                            String message =
//                                    "Error registering renewed sliver(s) at " + rspecComponentManager + ". This is not critical, and will be ignored. This will not influence the exit code.";
//                            if (e instanceof JFedException) {
//                                JFedException je = (JFedException) e;
//                                if (je.getXmlRpcResult() != null)
//                                    err.println(message + ": " + "\"" + je.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                                else {
//                                    err.println(message + ":");
//                                    e.printStackTrace(err);
//                                }
//                            } else {
//                                err.println(message + ":");
//                                e.printStackTrace(err);
//                            }
//                        }
//                }
//            }
//
//            if (cli.isAction(MANIFEST, CREATE)) {
//                if (cli.debug || (!cli.silent && cli.isCommand(MANIFEST)))
//                    out.println("Writing combined manifest to file \"" + cli.manifestFilename + "\"");
//                String combinedManifestXmlString = manifestRspecMerger.getMergedManifest().getRspecXmlString();
//                IOUtils.stringToFile(new File(cli.manifestFilename), combinedManifestXmlString);
//
//                if (cli.ansibleDir != null) {
//                    if (!cli.ansibleDir.exists()) {
//                        boolean created = cli.ansibleDir.mkdir();
//                        if (!cli.ansibleDir.exists()) {
//                            err.println("ERROR: Could not create ansible dir: \""+cli.ansibleDirName+"\"");
//                        }
//                    }
//                    if (cli.ansibleDir.exists()) {
//                        AnsibleFileWriter writer = null;
//                        if (user.getPrivateKeyFile() != null) {
//                            try {
//                                String privKey = IOUtils.fileToString(user.getPrivateKeyFile());
//                                if (!KeyUtil.isPemPrivateKeyEncrypted(privKey)) {
//                                    writer = AnsibleFileWriter.createWithLinkedPrivateKey(
//                                            new BasicStringRspec(combinedManifestXmlString),
//                                            user.getPrivateKeyFile(),
//                                            user.getPublicKey(),
//                                            user,
//                                            null/*sshProxyInfo*/,
//                                            null /*preferredUser*/);
//                                }
//                            } catch (Exception e) {
//                                LOG.error("Exception trying to determine if private key file is encrypted. Safe fallback: Will assume it is.", e);
//                                writer = null;
//                            }
//                        }
//                        if (writer == null) {
//                            writer = AnsibleFileWriter.createWithCopiedPrivateKey(
//                                    new BasicStringRspec(combinedManifestXmlString),
//                                    user.getPrivateKey(),
//                                    user.getPublicKey(),
//                                    user,
//                                    null, null /*preferredUser*/);
//                        }
//                        writer.writeFilesToDir(cli.ansibleDir);
//                    }
//                }
//            }
//
//            if (cli.isCommand(STATUS)) {
//                if (earliestSliverExpireDate != null)
//                    outerStatusHt.put("earliestSliverExpireDate", RFC3339Util.dateToRFC3339String(earliestSliverExpireDate));
//                if (cli.isJson && outerStatusHt != null) {
//                    out.println(JSonHelper.jsonObjectToString(outerStatusHt));
//                }
//                if (cli.isTextOutput && outerStatusHt != null) {
//                    out.println("Status Info:");
//                    out.println("   Slice URN: "+outerStatusHt.get("sliceUrn"));
//                    out.println("   Slice expire date: "+outerStatusHt.get("sliceExpireDate"));
//                    if (outerStatusHt.get("earliestSliverExpireDate") != null)
//                        out.println("   Earliest sliver expire date: "+outerStatusHt.get("earliestSliverExpireDate"));
//                    else
//                        out.println("   Earliest sliver expire date: no slivers");
//                    Map amsHt = (Map) outerStatusHt.get("AMs");
//                    Set<Map.Entry> entries = amsHt.entrySet();
//                    for (Map.Entry e : entries) {
//                        String key = (String) e.getKey();
//                        Map amHt = (Map)e.getValue();
//
//                        String amUrn = amHt.get("amUrn")+"";
//                        String amGlobalSliverStatus = amHt.get("amGlobalSliverStatus")+"";
//                        String earliestAMSliverExpireDateStr = amHt.get("earliestSliverExpireDate") == null ? null : amHt.get("earliestSliverExpireDate")+"";
//
//                        out.println("   AM: "+amUrn+" "+amGlobalSliverStatus+(earliestAMSliverExpireDateStr == null ? "" : " "+earliestAMSliverExpireDateStr));
//
//                        List details= (List) amHt.get("details");
//                        for (Object detail : details) {
//                            Map detailHt = (Map) detail;
//                            String urn = detailHt.get("urn")+"";
//                            String type = (String)detailHt.get("type");
//                            String status = detailHt.get("status")+"";
//                            out.println("      "+type+": "+urn+" "+status);
//                        }
//                    }
//                }
//            }
//
//            if (cli.isCommand(CREATE)) {
//                if (cli.waitUntilReady) {
//                    if (!cli.silent)
//                        if (rspecComponentManagers.size() == 1)
//                            out.println("Will now wait until the sliver is ready...");
//                        else
//                            out.println("Will now wait until the " + rspecComponentManagers.size() + " slivers are ready...");
//
//                    for (AggregateManagerWrapper am : amWrappers) {
//                        Server currentAuthority = am.getAmServer();
//                        String rspecComponentManager = currentAuthority.getDefaultComponentManagerUrn();
//
//                        StatusDetails statusDetails = null;
//                        while (statusDetails == null || !Objects.equals(statusDetails.getGlobalStatus(), StatusDetails.SliverStatus.READY)) {
//                            if (cli.debug)
//                                out.println("Waiting 5 seconds before checking status...");
//                            try {
//                                Thread.sleep(5 * 1000L);
//                            } catch (InterruptedException e) {
//                                LOG.debug("InterruptedException while sleeping", e);
//                            }
//
//                            if (!cli.silent)
//                                out.println("Contacting " + rspecComponentManager + " to check status...");
//
//                            try {
//                                statusDetails = am.status(usedSliceUrn, sliceCredentials);
//                            } catch (JFedException e) {
//                                if (cli.deleteOnBecomeReadyFailure)
//                                    deleteAll(amWrappers, cli.silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
//                                return callFailed("Error getting sliver status at " + currentAuthority.getDefaultComponentManagerUrn(), e, err, callLogOutput);
//                            }
//
//                            if (!cli.silent)
//                                out.println("Status of sliver at  " + rspecComponentManager + " is " + statusDetails.getGlobalStatus());
//
//                            if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL) ||
//                                    statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNALLOCATED)) {
//                                if (cli.deleteOnBecomeReadyFailure)
//                                    deleteAll(amWrappers, cli.silent, out, err, usedSliceUrn, sliceCredentials, logger, sa);
//                                return callFailed("ERROR: Sliver status at " + currentAuthority.getDefaultComponentManagerUrn() + " is " +
//                                        statusDetails.getGlobalStatus() + " -> " + statusDetails, null, err, callLogOutput);
//                            }
//                        }
//                    }
//
//                    if (!lanToShare.isEmpty()) {
//                        LOG.debug("Will try to share LANs");
//
//                        if (!cli.silent)
//                            out.println((lanToShare.size())+" LAN(s) need to be shared. Will wait 10 seconds before trying.");
//                        try {
//                            Thread.sleep(10);
//                        } catch (InterruptedException e) {
//                            //ignore
//                        }
//
//                        for (AggregateManagerWrapper am : amWrappers) {
//                            Server currentAuthority = am.getAmServer();
//
//                            for (ExperimenterCli.AutoShareLanInfo info : lanToShare) {
//                                if (Objects.equals(currentAuthority.getDefaultComponentManagerAsGeniUrn(), info.componentManagerUrn)) {
//                                    LOG.debug("Will try to share LAN "+info.lanClientId+" as "+info.sharedLanName+" on "+info.componentManagerUrn);
//                                    if (am.isShareLanSupported()) {
//                                        try {
//                                            am.shareLan(usedSliceUrn, sliceCredentials, info.lanClientId, info.sharedLanName);
//
//                                            if (!cli.silent) {
//                                                out.println("Shared LAN " + info.lanClientId + " as " + info.sharedLanName + " on " + info.componentManagerUrn);
//                                            }
//                                        } catch (JFedException e) {
//                                            LOG.error("Failed to share LAN", e);
//                                            err.println("Failed to share LAN " + info.lanClientId + " as " + info.sharedLanName + " on " + info.componentManagerUrn + ": "+e.getMessage());
//                                            e.printStackTrace();
//                                        }
//                                    } else {
//                                        err.println("Cannot share LAN "+info.lanClientId+" as "+info.sharedLanName+" on "+info.componentManagerUrn+": AM does not support share LAN");
//                                    }
//                                }
//                            }
//
//                        }
//                    }
//
//                    if (!cli.silent)
//                        if (rspecComponentManagers.size() == 1)
//                            out.println("The sliver is ready.");
//                        else
//                            out.println("The " + rspecComponentManagers.size() + " slivers are ready.");
//                }
//            }
//
//            if (!ansiblePlaybooksToExecute.isEmpty()) {
//                LOG.debug("Will execute "+ansiblePlaybooksToExecute.size()+" Ansible Playbooks");
//
//                if (!cli.silent)
//                    out.println((ansiblePlaybooksToExecute.size())+" ansible playbooks will be executed. Will wait 5 seconds before beginning.");
//                try {
//                    Thread.sleep(5);
//                } catch (InterruptedException e) {
//                    //ignore
//                }
//
//                int playbookIndex = 0;
//                for (ExperimenterCli.AnsiblePlaybookInfo info : ansiblePlaybooksToExecute) {
//                    File outputFile;
//                    if (info.getExecuteAnsiblePlaybook().getOutputFilename() == null)
//                        outputFile = new File(cli.ansibleDir, "playbook-"+playbookIndex+"-output.txt");
//                    else
//                        outputFile = cli.ansibleAllowAnyPlaybookOutputFile ? new File(info.getOutputFilename()) : new File(cli.ansibleDir, info.getOutputBasename());
//                    File playbookFile = new File(cli.ansibleDir, "playbook-"+playbookIndex+".yml");
//                    try {
//                        IOUtils.stringToFile(playbookFile, info.getScriptContent());
//
//                        if (!cli.silent)
//                            out.println("Executing ansible playbook \""+playbookFile.getName()+"\"...");
//                        callAnsible(cli.ansiblePlayBookExe, playbookFile, cli.ansibleDir, cli.ansibleDebug, outputFile);
//                    } catch (Exception e) {
//                        successExitStatus = false;
//                        err.println("Exception while executing ansible playbook (will still try other ansible playbooks, if any, but exit status will indicate failure). Exception:");
//                        e.printStackTrace();
//                    }
//
//                    playbookIndex++;
//                }
//            }


        LOG.debug("All done, exiting.");
    }

//    private static void callAnsible(String ansiblePlaybookExe, File ansiblePlaybook, File ansibleDir, boolean ansibleDebug, File outputFile) throws IOException {
//        PrintWriter pw = new PrintWriter(new FileWriter(outputFile));
//        ArrayList<String> command = new ArrayList<>();
//        assert ansiblePlaybookExe != null;
//        String[] ansiblePlaybookExeParts = ansiblePlaybookExe.split(" ");
//        Collections.addAll(command, ansiblePlaybookExeParts);
//        command.add(ansiblePlaybook.getAbsolutePath());
//        if (ansibleDebug)
//            command.add("-vvvv");
//        ProcessBuilder pb = new ProcessBuilder(command);
//        pb.directory(ansibleDir);
//        pb.redirectErrorStream(); //combine stderr and stout, so we can read both at once in one thread. Makes everything easy.
//
//        Process p = pb.start();
//        BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
//
////        String output = "";
//        String line;
//        while ((line = input.readLine()) != null) {
////            output += line + System.lineSeparator();
//            pw.println(line);
//        }
//        //note: input might close before terminal stops.
//        input.close();
//
//        try {
//            p.waitFor();
//        } catch (InterruptedException e) {
//            //ignore
//        }
//
//        pw.close();
//    }


//    private static void deleteAll(List<AggregateManagerWrapper> amWrappers, boolean silent,
//                                  PrintStream out, PrintStream err, GeniUrn usedSliceUrn,
//                                  List<AnyCredential> sliceCredentials, be.iminds.ilabt.jfed.log.Logger logger,
//                                  UserAndSliceApiWrapper sa) {
//        if (!silent)
//            err.println("Failure: will delete all slivers, on "+amWrappers+" AMs.");
//
//        for (AggregateManagerWrapper am : amWrappers) {
//            Server currentAuthority = am.getAmServer();
//            String rspecComponentManager = currentAuthority.getDefaultComponentManagerUrn();
//            GeniUrn rspecComponentManagerUrn = currentAuthority.getDefaultComponentManagerAsGeniUrn();
//            if (!silent)
//                out.println("Contacting  " + rspecComponentManager + "...");
//
//            try {
//                am.deleteSliver(usedSliceUrn, sliceCredentials);
//
//                if (!silent)
//                    out.println("Slivers at  " + rspecComponentManager + " have been deleted.");
//
//                try {
//                    sa.unregisterAggregatesForSlice(logger, sliceCredentials, usedSliceUrn, rspecComponentManagerUrn, null/*sliverUrns*/);
//
//                    if (!silent)
//                        out.println("Slivers at  " + rspecComponentManager + " has been unregistered.");
//                } catch (JFedException e) {
//                    String message =
//                            "Error unregistering sliver at " + rspecComponentManager + ". (this error will be ignored)";
//                    if (e.getXmlRpcResult() != null)
//                        err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                    else {
//                        err.println(message + ":");
//                        e.printStackTrace(err);
//                    }
//                }
//            } catch (JFedException e) {
//                String message =
//                        "Error deleting sliver at " + rspecComponentManager + ". (this error will be ignored)";
//                if (e.getXmlRpcResult() != null)
//                    err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//                else {
//                    err.println(message + ":");
//                    e.printStackTrace(err);
//                }
//            }
//        }
//    }

//    private static int callFailed(String message, JFedException e, PrintStream err, ExperimenterCli.CallLogOutput callLogOutput) {
//        Thread.yield();
//        Thread.yield();
//        callLogOutput.flush();
//        try {
//            Thread.sleep(100);
//        } catch (InterruptedException e1) {
//            //ignored
//        }
//        Thread.yield();
//        callLogOutput.flush();
//
//        if (e == null)
//            err.println(message);
//        else {
//            if (e.getXmlRpcResult() != null)
//                err.println(message + ": " + "\"" + e.getXmlRpcResult().getResultXmlRpcString() + "\"");
//            else {
//                err.println(message + ":");
//                e.printStackTrace(err);
//            }
//        }
//        return 2;
//    }


    @Nonnull
    private JFedGuiConfig getGuiConfig(@Nonnull Injector injector) {
        final GeniUserProvider geniUserProvider = injector.getInstance(GeniUserProvider.class);
        final be.iminds.ilabt.jfed.log.Logger logger = injector.getInstance(be.iminds.ilabt.jfed.log.Logger.class);
        final TestbedInfoSource testbedInfoSource = injector.getInstance(TestbedInfoSource.class);
        final AuthorityFinder authorityFinder = injector.getInstance(AuthorityFinder.class);
        JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);
        ConnectivityDetector connectivityDetector = injector.getInstance(ConnectivityDetector.class);
        UserInfoProvider userInfoProvider = injector.getInstance(UserInfoProvider.class);
        URL webApiUrl =
                injector.getInstance(Key.get(URL.class, Names.named(JFedWebApiClientModule.NAMED__NO_CLIENT_AUTH_WEBAPI_URL)));
        FedmonWebApiClient fedmonWebApiClient = injector.getInstance(FedmonWebApiClient.class);

        if (geniUserProvider == null) {
            throw new NullPointerException("geniUserProvider may not be null");
        }

        JFedExperimenterGuiConfigProvider jFedExperimenterGuiConfigProvider =
                new JFedExperimenterGuiConfigProvider(
                        geniUserProvider,
                        jFedPreferences,
                        connectivityDetector,
                        userInfoProvider,
                        webApiUrl,
                        logger,
                        false);
        TestbedNodesMapsFetcher testbedNodesMapFetcher =
                new TestbedNodesMapsFetcher(connectivityDetector);
        JFedGuiConfigImpl res = new JFedGuiConfigImpl(
                testbedInfoSource,
                authorityFinder,
                jFedExperimenterGuiConfigProvider,
                testbedNodesMapFetcher,
                fedmonWebApiClient);
        return res;
    }

    @Nonnull
    private RSpecGeneratorFactory getRSpecGeneratorFactory(@Nonnull Injector injector) {
        final TestbedInfoSource testbedInfoSource = injector.getInstance(TestbedInfoSource.class);
        final AuthorityFinder authorityFinder = injector.getInstance(AuthorityFinder.class);

        return new RSpecGeneratorFactory() {
            @Nonnull
            @Override
            public RSpecGenerator createRSpecGenerator() {
                return new GuiConfigRSpecGenerator(
                        getGuiConfig(injector),
                        authorityFinder,
                        testbedInfoSource);
            }

            @Nonnull
            @Override
            public StitchingTestRSpecGenerator createStitchingTestRSpecGenerator() {
                return new GuiConfigStitchingTestRSpecGenerator(
                        getGuiConfig(injector),
                        authorityFinder,
                        testbedInfoSource);
            }
        };
    }


    private List<UserSpec> getUserSpecs(RunInstruction.SshKeySource sshKeySource) {
        List<UserSpec> res = new ArrayList<>();
        if (sshKeySource.isUserCert()) {
            List<String> sshKeys = Collections.singletonList(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(context.getUser().getPublicKey()));
            UserSpec userSpec = new UserSpec(context.getUser().getUserUrnString(), sshKeys);
            res.add(userSpec);
        }
        if (sshKeySource.isUserKeys()) {
            //TODO
        }
        if (sshKeySource.isShareWith()) {
            //TODO
        }
        if (sshKeySource.isRspec()) {
            //TODO
        }
        if (!sshKeySource.getExtraKeys().isEmpty()) {
            //TODO
        }
        if (!sshKeySource.getExtraUsernames().isEmpty()) {
            //TODO
        }
        return res;
    }

    private static GeniUrn verifyBindRspecUrn(
            @Nonnull Context context,
            @Nonnull String bindRspecArg,
            @Nonnull AuthorityFinder authorityFinder) throws CliExitException {
        GeniUrn bindRspecUrn = GeniUrn.parse(bindRspecArg);
        if (bindRspecUrn == null) {
            context.exitWithCode(Context.ExitReason.BAD_ARGS, "Error parsing provided component_manager_id URN: " + bindRspecArg);
            return null;
        }
        Server bindRspecAuth = authorityFinder.findByUrn(bindRspecUrn, AuthorityFinder.Purpose.REQUEST_RSPEC);
        if (bindRspecAuth == null) {
            context.exitWithCode(Context.ExitReason.BAD_ARGS, "Error: could not find server matching component_manager_id URN provided by --bind-rspec: " + bindRspecArg);
            return null;
        }
        return bindRspecUrn;
    }

    private static class ESpecLoggingToOutput implements ESpecLogListener {
        @Nonnull private final ActionLogger actionLogger;

        public ESpecLoggingToOutput(@Nonnull ActionLogger actionLogger) {
            this.actionLogger = actionLogger;
        }

        @Override
        public void onPreFileLoad() {
            actionLogger.actionPrintln("ExperimentSpecification : Start loading files.");
        }

        @Override
        public void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize,
                                   @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                actionLogger.errorPrintln("ExperimentSpecification : Failed to load file \""+fileSource.getBasename()+"\"",
                        status.getThrowable());
            else {
                if (!isFast)
                    actionLogger.actionPrintln("ExperimentSpecification : Successfully loaded file \""+fileSource.getBasename()+"\" ("+byteSize+" byte)");
            }
        }

        @Override
        public void onPostFileLoadAll(boolean success) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to load all files.");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Successfully finished loading files.");
        }

        @Override
        public void onPreRSpec() {
            actionLogger.actionPrintln("ExperimentSpecification : Start Provisioning RSpec.");
        }

        @Override
        public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
            actionLogger.actionPrintln("ExperimentSpecification : Request RSpec known:\n"+requestRspec);
        }

        @Override
        public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                actionLogger.errorPrintln("Failed setting up experiment resources", status.getThrowable());
        }

        @Override
        public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to Provision RSpec");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully Provisioned RSpec with nodes "+allNodeClientIds.stream().collect(Collectors.joining()));
        }

        @Override
        public void onPreDir() {
            actionLogger.actionPrintln("ExperimentSpecification : Start dir setup.");
        }

        @Override
        public void onPreDirNode(@Nonnull String nodeClientId) {

        }

        @Override
        public void onPreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId) {

        }

        @Override
        public void onPostDir(long logEventId,
                              @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId,
                              @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                actionLogger.errorPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") " +
                        "Failed to setup dir at \""+fullPath+"\"", status.getThrowable());
            else
                actionLogger.actionPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") Succesfully setup dir at \""+fullPath+"\"");
        }

        @Override
        public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {

        }

        @Override
        public void onPostDirAll(boolean success) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to setup dirs");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully setup dirs");
        }

        @Override
        public void onPreUpload() {
            actionLogger.actionPrintln("ExperimentSpecification : Start upload(s).");
        }

        @Override
        public void onPreUploadNode(@Nonnull String nodeClientId) {

        }

        @Override
        public void onPreUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull UploadProgressTracker uploadProgressTracker) {

        }

        @Override
        public void onPostUpload(long logEventId,
                                 @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                actionLogger.errorPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") Failed to upload file \""+uploadSpec.getDesc()+"\" at \""+fullPath+"\"");
            else {
                actionLogger.actionPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") Succesfully uploaded file \"" + uploadSpec.getDesc() + "\" at \"" + fullPath + "\"");
            }
        }

        @Override
        public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {

        }

        @Override
        public void onPostUploadAll(boolean success) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to upload at least one file to one node");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully uploaded files");
        }

        @Override
        public void onPreExecute() {
            actionLogger.actionPrintln("ExperimentSpecification : Start execute(s).");
        }

        @Override
        public void onPreExecute(long logEventId, @Nonnull ExecuteSpec executeSpec, @Nonnull String fullExecutablePath, @Nullable String fullLogPath, @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) {

        }

        @Override
        public void onPostExecute(long logEventId,
                                  @Nonnull ExecuteSpec executeSpec,
                                  @Nonnull String fullPath,
                                  @Nullable String fullLogPath,
                                  @Nonnull String nodeClientId,
                                  int exitStatus,
                                  @Nonnull ESpecStepStatus status) {
            if (status.isFailure())
                actionLogger.errorPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") " +
                        "Failed to execute script \"" + executeSpec.getDesc() + "\" at \""+fullPath+"\" " +
                        "exitStatus="+exitStatus, status.getThrowable());
            else
                actionLogger.actionPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") " +
                        "Succesfully executed script \"" + executeSpec.getDesc() + "\" at \""+fullPath+"\" " +
                        "exitStatus="+exitStatus);
        }

        @Override
        public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to execute script \"" + executeSpec.getDesc() + "\" on at least one node");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully executed script \"" + executeSpec.getDesc() + "\" on all nodes");
        }

        @Override
        public void onPostExecuteAll(boolean success) {
            if (!success)
                actionLogger.errorPrintln("ExperimentSpecification : Failed to execute at least one script on at least one node");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully executed all scripts");
        }

        @Override
        public void onPreAnsible() {
            actionLogger.actionPrintln("ExperimentSpecification : Start ansible(s).");
        }

        @Override
        public void onPreAnsiblePlaybook(long logEventId, @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec, @Nonnull String fullPlaybookPath, @Nullable String fullLogPath, @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) {

        }

        @Override
        public void onPostAnsiblePlaybook(long logEventId,
                                          @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                          @Nonnull String fullPath,
                                          @Nullable String fullLogPath,
                                          @Nonnull String nodeClientId,
                                          @Nonnull ESpecStepStatus status) {
            if (status.isFailure()) {
                actionLogger.actionPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") " +
                        "Failed to execute ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" " +
                        "at \"" + fullPath + "\""+
                        (status.getThrowable() == null ? "" : ": "+ IOUtils.exceptionToStacktraceString(status.getThrowable())));
            } else
                actionLogger.actionPrintln("ExperimentSpecification : (on \"" + nodeClientId + "\") " +
                        "Succesfully executed ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" " +
                        "at \""+fullPath+"\"");
        }

        @Override
        public void onPostAnsibleAll(boolean success) {
            if (!success)
                /*errorNonFatal*/actionLogger.actionPrintln("ExperimentSpecification : Failed to ansible at least one script on at least one node");
            else
                actionLogger.actionPrintln("ExperimentSpecification : Succesfully ansibled all scripts");
        }
    }

    private static class ESpecLoggingCounter implements ESpecLogListener {
        @Nonnull private Map<String, Integer> counts = new HashMap<>();

        private void inc(@Nonnull String type, boolean failure) {
            String key = type + (failure ? "-fail" :  "-ok");
            counts.compute(key, (k, v) -> v == null ? 1 : v + 1);
        }

        public ESpecLoggingCounter() {
        }

        @Override public void onPreFileLoad() { }

        @Override
        public void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize,
                                   @Nonnull ESpecStepStatus status) {
            inc("upload", status.isFailure());
        }

        @Override public void onPostFileLoadAll(boolean success) { }

        @Override public void onPreRSpec() { }
        @Override public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) { }

        @Override
        public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {
            inc("rspec", status.isFailure());
        }

        @Override public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) { }

        @Override public void onPreDir() { }
        @Override public void onPreDirNode(@Nonnull String nodeClientId) { }
        @Override public void onPreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId) { }

        @Override
        public void onPostDir(long logEventId,
                              @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId,
                              @Nonnull ESpecStepStatus status) {
            inc("dir", status.isFailure());
        }

        @Override public void onPostDirNode(@Nonnull String nodeClientId, boolean success) { }
        @Override public void onPostDirAll(boolean success) { }

        @Override public void onPreUpload() { }
        @Override public void onPreUploadNode(@Nonnull String nodeClientId) { }
        @Override public void onPreUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull UploadProgressTracker uploadProgressTracker) { }

        @Override
        public void onPostUpload(long logEventId,
                                 @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
            inc("upload", status.isFailure());
        }

        @Override public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) { }
        @Override public void onPostUploadAll(boolean success) { }

        @Override public void onPreExecute() { }
        @Override public void onPreExecute(long logEventId, @Nonnull ExecuteSpec executeSpec, @Nonnull String fullExecutablePath, @Nullable String fullLogPath, @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) { }

        @Override
        public void onPostExecute(long logEventId,
                                  @Nonnull ExecuteSpec executeSpec,
                                  @Nonnull String fullPath,
                                  @Nullable String fullLogPath,
                                  @Nonnull String nodeClientId,
                                  int exitStatus,
                                  @Nonnull ESpecStepStatus status) {
            inc("execute", status.isFailure());
        }

        @Override public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) { }
        @Override public void onPostExecuteAll(boolean success) { }

        @Override public void onPreAnsible() { }
        @Override public void onPreAnsiblePlaybook(long logEventId, @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec, @Nonnull String fullPlaybookPath, @Nullable String fullLogPath, @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) { }

        @Override
        public void onPostAnsiblePlaybook(long logEventId,
                                          @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                          @Nonnull String fullPath,
                                          @Nullable String fullLogPath,
                                          @Nonnull String nodeClientId,
                                          @Nonnull ESpecStepStatus status) {
            inc("ansible", status.isFailure());
        }

        @Override public void onPostAnsibleAll(boolean success) { }

        @Nonnull
        public Map<String, Integer> getCounts() {
            return counts;
        }
    }

    private static class LinkTestResultToOutput implements LinkTestListener {
        @Nonnull private final ActionLogger actionLogger;

        @Nullable private List<TestLinksJob.LinkTestReport> reports;
        @Nullable private TestLinksJob.LinkTestResult worstResult;

        public LinkTestResultToOutput(@Nonnull ActionLogger actionLogger) {
            this.actionLogger = actionLogger;
        }

        @Override
        public void onReport(@Nonnull TestLinksJob.LinkTestReport report) {
            //don't report individually
        }

        @Override
        public void onAllReports(@Nonnull List<TestLinksJob.LinkTestReport> reports,
                                 @Nonnull TestLinksJob.LinkTestResult worstResult) {
            this.reports = reports;
            this.worstResult = worstResult;

            String reportSummaryString = reports.stream()
                    .map(TestLinksJob.LinkTestReport::toStringNoSuccessDebug)
                    .collect(Collectors.joining("\n"));
            if (worstResult == TestLinksJob.LinkTestResult.SUCCESS) {
                actionLogger.actionPrintln("Link Test Result: "+worstResult);
                LOG.debug("Link Test Reports:\n" + reportSummaryString);
            } else {
                actionLogger.errorPrintln("Link Test Result: "+worstResult);
                actionLogger.actionPrintln("Link Test Reports:\n" + reportSummaryString);
            }
        }

        @Nullable
        public List<TestLinksJob.LinkTestReport> getReports() {
            return reports;
        }

        @Nullable
        public TestLinksJob.LinkTestResult getWorstResult() {
            return worstResult;
        }
    }
}
